from tests.base_test_case import BaseTestCase

from orquestaconvert.expressions.jinja import JinjaExpressionConverter


class TestExpressionsJinja(BaseTestCase):
    __test__ = True

    def test_unwrap_expression(self):
        expr = "{{ _.test }}"
        result = JinjaExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, "_.test")

    def test_unwrap_expression_nested(self):
        expr = "{{ _.test {{ abc }} }}"
        result = JinjaExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, "_.test {{ abc }}")

    def test_unwrap_expression_trim_spaces(self):
        expr = "{{           _.test       }}"
        result = JinjaExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, "_.test")

    def test_convert_expression_jinja_context_vars(self):
        expr = "{{ _.test }}"
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ ctx().test }}")

    def test_convert_expression_jinja_item_vars(self):
        expr = "{{ _.test }}"
        result = JinjaExpressionConverter.convert_string(expr, item_vars=['test'])
        self.assertEqual(result, "{{ item(test) }}")

    def test_convert_expression_jinja_context_and_item_vars(self):
        expr = "{{ _.test + _.test2 - _.long_var }}"
        result = JinjaExpressionConverter.convert_string(expr, item_vars=['test'])
        self.assertEqual(result, "{{ item(test) + ctx().test2 - ctx().long_var }}")

    def test_convert_expression_jinja_function_context_vars(self):
        expr = "{{ list(range(0, _.count)) }}"
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ list(range(0, ctx().count)) }}")

    def test_convert_expression_jinja_complex_function_context_vars(self):
        expr = "{{ zip([0, 1, 2], [3, 4, 5], _.all_the_things) }}"
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ zip([0, 1, 2], [3, 4, 5], ctx().all_the_things) }}")

    def test_convert_expression_jinja_context_vars_multiple(self):
        expr = "{{ _.test + _.other }}"
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ ctx().test + ctx().other }}")

    def test_convert_expression_jinja_context_vars_with_underscore(self):
        expr = "{{ _.test_.other }}"
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ ctx().test_.other }}")

    def test_convert_expression_jinja_task_result(self):
        expr = "{{ task('abc').result.result }}"
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ result().result }}")

    def test_convert_expression_jinja_task_result_double_quotes(self):
        expr = '{{ task("abc").result.double_quote }}'
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ result().double_quote }}")

    def test_convert_expression_jinja_st2kv(self):
        expr = '{{ st2kv.system.test.kv }}'
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ st2kv('system.test.kv') }}")

    def test_convert_expression_jinja_st2kv_user(self):
        expr = '{{ st2kv.user.test.kv }}'
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ st2kv('user.test.kv') }}")

    def test_convert_expression_jinja_st2_execution_id(self):
        expr = '{{ env().st2_execution_id }}'
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ ctx().st2.action_execution_id }}")

    def test_convert_expression_jinja_st2_api_url(self):
        expr = '{{ env().st2_action_api_url }}'
        result = JinjaExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ ctx().st2.api_url }}")



import re

from tests.base_test_case import BaseTestCase

from orquestaconvert.expressions import ExpressionConverter
from orquestaconvert.expressions.jinja import JinjaExpressionConverter
from orquestaconvert.expressions.yaql import YaqlExpressionConverter


class TestExpressions(BaseTestCase):
    __test__ = True

    def test_convert_expr_bool(self):
        expr_bool = True
        result = ExpressionConverter.convert(expr_bool)
        self.assertEqual(result, True)
        expr_bool = False
        result = ExpressionConverter.convert(expr_bool)
        self.assertEqual(result, False)

    def test_convert_expr_null(self):
        expr_none = None
        result = ExpressionConverter.convert(expr_none)
        self.assertEqual(result, None)

    def test_convert_expr_int(self):
        expr_int = 0
        result = ExpressionConverter.convert(expr_int)
        self.assertEqual(result, 0)
        expr_int = 100
        result = ExpressionConverter.convert(expr_int)
        self.assertEqual(result, 100)
        expr_int = -1
        result = ExpressionConverter.convert(expr_int)
        self.assertEqual(result, -1)

    def test_convert_expr_obj_with_warning(self):
        expr_obj = object()
        expected_warning_regex = re.compile(
            r"Could not recognize expression '<object object at 0x[0-9a-f]+>'; "
            r"results may not be accurate.")
        with self.assertWarnsRegex(SyntaxWarning, expected_warning_regex):
            result = ExpressionConverter.convert(expr_obj)
        self.assertEqual(result, expr_obj)

    def test_convert_expr_dict(self):
        expr_dict = {"test": "{{ _.value }}"}
        result = ExpressionConverter.convert(expr_dict)
        self.assertEqual(result, {"test": "{{ ctx().value }}"})

    def test_convert_expr_list(self):
        expr_list = ["test", "{{ _.value }}"]
        result = ExpressionConverter.convert(expr_list)
        self.assertEqual(result, ["test", "{{ ctx().value }}"])

    def test_convert_expr_string(self):
        expr_dict = "{{ _.value }}"
        result = ExpressionConverter.convert(expr_dict)
        self.assertEqual(result, "{{ ctx().value }}")

    def test_convert_no_expression(self):
        expr = "data"
        result = ExpressionConverter.convert(expr)
        self.assertEqual(result, "data")

    def test_expression_type_jinja(self):
        expr = "{{ _.test }}"
        result = ExpressionConverter.expression_type(expr)
        self.assertEqual(result, 'jinja')

    def test_expression_type_yaql(self):
        expr = "<% $.test %>"
        result = ExpressionConverter.expression_type(expr)
        self.assertEqual(result, 'yaql')

    def test_expression_type_none(self):
        expr = "test"
        result = ExpressionConverter.expression_type(expr)
        self.assertIsNone(result)

    def test_get_converter_jinja(self):
        expr = "{{ _.test }}"
        result = ExpressionConverter.get_converter(expr)
        self.assertIs(result, JinjaExpressionConverter)

    def test_get_converter_yaql(self):
        expr = "<% $.test %>"
        result = ExpressionConverter.get_converter(expr)
        self.assertIs(result, YaqlExpressionConverter)

    def test_get_converter_none(self):
        expr = "test"
        result = ExpressionConverter.get_converter(expr)
        self.assertIsNone(result)

    def test_unwrap_expression_jinja(self):
        expr = "{{ _.test }}"
        result = ExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, '_.test')

    def test_unwrap_expression_yaql(self):
        expr = "<% $.test %>"
        result = ExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, '$.test')

    def test_unwrap_expression_none(self):
        expr = "test"
        result = ExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, 'test')

    def test_convert_dict(self):
        expr = {
            "jinja_str": "{{ _.test_jinja }}",
            "yaql_str": "<% $.test_yaql %>",
        }
        result = ExpressionConverter.convert_dict(expr)
        self.assertEqual(result, {
            "jinja_str": "{{ ctx().test_jinja }}",
            "yaql_str": "<% ctx().test_yaql %>",
        })

    def test_convert_dict_nested_list(self):
        expr = {
            "expr_list":
            [
                "{{ _.a }}",
                "<% $.a %>",
            ]
        }
        result = ExpressionConverter.convert_dict(expr)
        self.assertEqual(result, {
            "expr_list":
            [
                "{{ ctx().a }}",
                "<% ctx().a %>",
            ]
        })

    def test_convert_dict_nested_dict(self):
        expr = {
            "expr_dict":
            {
                "nested_jinja": "{{ _.a }}",
                "nested_yaql": "<% $.a %>",
            }
        }
        result = ExpressionConverter.convert_dict(expr)
        self.assertEqual(result, {
            "expr_dict":
            {
                "nested_jinja": "{{ ctx().a }}",
                "nested_yaql": "<% ctx().a %>",
            }
        })

    def test_convert_list(self):
        expr = [
            "{{ _.test_jinja }}",
            "<% $.test_yaql %>",
        ]
        result = ExpressionConverter.convert_list(expr)
        self.assertEqual(result, [
            "{{ ctx().test_jinja }}",
            "<% ctx().test_yaql %>",
        ])

    def test_convert_list_nested_list(self):
        expr = [
            "{{ _.a }}",
            [
                "<% $.a %>",
            ]
        ]
        result = ExpressionConverter.convert_list(expr)
        self.assertEqual(result, [
            "{{ ctx().a }}",
            [
                "<% ctx().a %>",
            ]
        ])

    def test_convert_list_nested_dict(self):
        expr = [
            {
                "nested_jinja": "{{ _.a }}",
                "nested_yaql": "<% $.a %>",
            }
        ]
        result = ExpressionConverter.convert_list(expr)
        self.assertEqual(result, [
            {
                "nested_jinja": "{{ ctx().a }}",
                "nested_yaql": "<% ctx().a %>",
            }
        ])

    def test_convert_string_other(self):
        expr = "test some raw string"
        result = ExpressionConverter.convert_string(expr)
        self.assertEqual(result, "test some raw string")

    def test_convert_string_jinja(self):
        expr = "{{ _.test }}"
        result = ExpressionConverter.convert_string(expr)
        self.assertEqual(result, "{{ ctx().test }}")

    def test_convert_string_yaql(self):
        expr = "<% $.test %>"
        result = ExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% ctx().test %>")



from tests.base_test_case import BaseTestCase

from orquestaconvert.expressions.yaql import YaqlExpressionConverter


class TestExpressionsYaql(BaseTestCase):
    __test__ = True

    def test_unwrap_expression(self):
        expr = "<% $.test %>"
        result = YaqlExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, "$.test")

    def test_unwrap_expression_nested(self):
        expr = "<% $.test <% abc %> %>"
        result = YaqlExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, "$.test <% abc %>")

    def test_unwrap_expression_trim_spaces(self):
        expr = "<%           $.test       %>"
        result = YaqlExpressionConverter.unwrap_expression(expr)
        self.assertEqual(result, "$.test")

    def test_convert_expression_yaql_context_vars(self):
        expr = "<% $.test %>"
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% ctx().test %>")

    def test_convert_expression_yaql_item_vars(self):
        expr = "<% $.test %>"
        result = YaqlExpressionConverter.convert_string(expr, item_vars=['test'])
        self.assertEqual(result, "<% item(test) %>")

    def test_convert_expression_yaql_context_and_item_vars(self):
        expr = "<% $.test + $.test2 - $.long_var %>"
        result = YaqlExpressionConverter.convert_string(expr, item_vars=['test'])
        self.assertEqual(result, "<% item(test) + ctx().test2 - ctx().long_var %>")

    def test_convert_expression_yaql_function_context_vars(self):
        expr = "<% list(range(0, $.count)) %>"
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% list(range(0, ctx().count)) %>")

    def test_convert_expression_yaql_complex_function_context_vars(self):
        expr = "<% zip([0, 1, 2], [3, 4, 5], $.all_the_things) %>"
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% zip([0, 1, 2], [3, 4, 5], ctx().all_the_things) %>")

    def test_convert_expression_yaql_context_vars_multiple(self):
        expr = "<% $.test + $.other %>"
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% ctx().test + ctx().other %>")

    def test_convert_expression_yaql_context_vars_with_underscore(self):
        expr = "<% $.test_.other %>"
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% ctx().test_.other %>")

    def test_convert_expression_yaql_task_result(self):
        expr = "<% task(abc).result.result %>"
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% result().result %>")

    def test_convert_expression_yaql_task_result_single_quote(self):
        expr = "<% task('abc').result.result %>"
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% result().result %>")

    def test_convert_expression_yaql_task_result_double_quotes(self):
        expr = '<% task("abc").result.double_quote %>'
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% result().double_quote %>")

    def test_convert_expression_yaql_st2kv(self):
        expr = '<% st2kv.system.test.kv %>'
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% st2kv('system.test.kv') %>")

    def test_convert_expression_yaql_st2kv_user(self):
        expr = '<% st2kv.user.test.kv %>'
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% st2kv('user.test.kv') %>")

    def test_convert_expression_yaql_st2_execution_id(self):
        expr = '<% env().st2_execution_id %>'
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% ctx().st2.action_execution_id %>")

    def test_convert_expression_yaql_st2_api_url(self):
        expr = '<% env().st2_action_api_url %>'
        result = YaqlExpressionConverter.convert_string(expr)
        self.assertEqual(result, "<% ctx().st2.api_url %>")



import re

from tests.base_test_case import BaseTestCase

from orquestaconvert.expressions.mixed import MixedExpressionConverter


class TestExpressionsMixed(BaseTestCase):
    __test__ = True

    def test_convert_string(self):
        s = "ABC {{ _.test }}"
        result = MixedExpressionConverter.convert(s)
        self.assertEqual(result, "ABC {{ ctx().test }}")

    def test_convert_string_items(self):
        s = "i in {{ _.test1 }}<% $.test2 %>{{ _.test3 }}<% $.test4 %>"
        result = MixedExpressionConverter.convert(s, item_vars=['test1', 'test2'])
        expected = "i in {{ item(test1) }}<% item(test2) %>{{ ctx().test3 }}<% ctx().test4 %>"
        self.assertEqual(result, expected)

    def test_convert_dict(self):
        data = {
            "test1": "FOO {{ _.bar }}",
            "test2": "FOOBAR <% $.baz %>",
        }
        expected = {
            "test1": "FOO {{ ctx().bar }}",
            "test2": "FOOBAR <% ctx().baz %>",
        }
        result = MixedExpressionConverter.convert(data)
        self.assertEqual(expected, result)

    def test_convert_list(self):
        data = [
            "FOO {{ _.bar }}",
            "FOOBAR <% $.baz %>",
        ]
        expected = [
            "FOO {{ ctx().bar }}",
            "FOOBAR <% ctx().baz %>",
        ]
        result = MixedExpressionConverter.convert(data)
        self.assertEqual(expected, result)

    def test_convert_dict_of_lists_of_dicts(self):
        data = {
            "list1": [
                {
                    "key1": "FOO1 {{ _.bar }}",
                    "key2": "FOO2 {{ _.baz }}",
                },
                {
                    "key3": "FOO3 <% $.car %>",
                    "key4": "FOO4 <% $.caz %>",
                },
            ],
            "int1": 1,
            "bool1": False,
            "null1": None,
        }
        expected = {
            "list1": [
                {
                    "key1": "FOO1 {{ ctx().bar }}",
                    "key2": "FOO2 {{ ctx().baz }}",
                },
                {
                    "key3": "FOO3 <% ctx().car %>",
                    "key4": "FOO4 <% ctx().caz %>",
                },
            ],
            "int1": 1,
            "bool1": False,
            "null1": None,
        }
        result = MixedExpressionConverter.convert(data)
        self.assertItemsEqual(expected.keys(), result.keys())
        self.assertItemsEqual(expected['list1'][0].keys(), result['list1'][0].keys())
        self.assertItemsEqual(expected['list1'][0].values(), result['list1'][0].values())
        self.assertEqual(expected['int1'], result['int1'])
        self.assertEqual(expected['bool1'], result['bool1'])
        self.assertEqual(expected['null1'], result['null1'])

    def test_convert_expr_obj_raises_warning(self):
        expr_obj = object()
        expected_warning_regex = re.compile(
            r"Could not recognize expression '<object object at 0x[0-9a-f]+>'; "
            r"results may not be accurate.")
        with self.assertWarnsRegex(SyntaxWarning, expected_warning_regex):
            result = MixedExpressionConverter.convert(expr_obj)
        self.assertEqual(result, expr_obj)



import ruamel.yaml
import ruamel.yaml.comments
import six
import yaml
import yamlloader


def yaml_to_obj(stream):
    return yaml.load(stream, Loader=yamlloader.ordereddict.CSafeLoader)


def read_yaml(yaml_filename):
    # parse data in a format that preserves ordering
    with open(yaml_filename, 'r') as stream:
        # safe load with ordered dicts
        ruamel_data = ruamel.yaml.round_trip_load(stream)

    # parse YAML into a dict
    with open(yaml_filename, 'r') as stream:
        data = yaml_to_obj(stream)

    return (data, ruamel_data)


def obj_to_yaml(obj, indent=2):
    # use this different library because PyYAML doesn't handle indenting properly
    # rt = round-trip
    ruyaml = ruamel.yaml.YAML(typ='rt')
    ruyaml.explicit_start = True
    # this crazyness basically sets indents to 'indent'
    # 'sequence' is always supposed to be 'offset' + 2
    ruyaml.indent(mapping=indent, sequence=(indent + 2), offset=indent)
    # prevent line-wrap
    ruyaml.width = 99999999999
    stream = six.StringIO()
    ruyaml.dump(obj, stream)
    return stream.getvalue()



import re
import six
import warnings

import ruamel.yaml

from orquestaconvert.expressions import ExpressionConverter
from orquestaconvert.expressions.base import BaseExpressionConverter
from orquestaconvert.expressions.jinja import JinjaExpressionConverter
from orquestaconvert.expressions.yaql import YaqlExpressionConverter
from orquestaconvert.expressions.mixed import MixedExpressionConverter
from orquestaconvert.utils import task_utils
from orquestaconvert.utils import type_utils

WORKFLOW_TYPES = [
    'direct',
]

WORKFLOW_UNSUPPORTED_ATTRIBUTES = [
    'output-on-error',
]

TASK_UNSUPPORTED_ATTRIBUTES = [
    'keep-result',
    'pause-before',
    'safe-rerun',
    'target',
    'timeout',
    'wait-after',
    'wait-before',
    'workflow',
]

UNSUPPORTED_MISTRAL_ACTIONS = [
    'std.echo',
    'std.email',
    'std.javascript',
    'std.js',
    'std.ssh',
]

MISTRAL_ACTION_CONVERSION_TABLE = {
    'std.fail': 'fail',
    'std.http': 'core.http',
    'std.mistral_http': 'core.http',
    'std.noop': 'core.noop',
}

# Comparison operators and their inverses (both Jinja and YAQL):
COMPARISON_OPERATOR_INVERSES = {
    '!=': '=',
    '<':  '>=',  # noqa: E241
    '<=': '>',
    '=':  '!=',  # noqa: E241
    '>':  '<=',  # noqa: E241
    '>=': '<',
    # YAQL has additional regex matching operators:
    '=~': '!~',
    '!~': '=~',
    # More importantly, Jinja does not have either of those operators, so this
    # should not inavertently any Jinja expressions
}


# This regex matches Mistral's with-items syntax:
# variable_name in <% YAQL_EXPRESSION %>
# variable_name in {{ JINJA_EXPRESSION }}
# variable_name in [static, list]
WITH_ITEMS_PATTERN = r'^\s*(?P<var>\w+)\s+in\s+(?P<expr>(?:<%|{{)?\s*.+?\s*(?:%>|}})?)\s*$'
WITH_ITEMS_EXPR_RGX = re.compile(WITH_ITEMS_PATTERN)

RETRY_BREAK_ON_PATTERN = r'^(?P<lhs>[^!~?=<>\s]+)\s*(?P<comp_op>!?=|<=?|>=?|[=!]~)\s*(?P<rhs>[^!~?=<>\s]+)$'
RETRY_BREAK_ON_RGX = re.compile(RETRY_BREAK_ON_PATTERN)

CTX_PATTERN = r'\bctx\([\'"]*(\w+)[\'"]*\)|\bctx\(\).(\w+)\b'
CTX_RGX = re.compile(CTX_PATTERN)


class WorkflowConverter(object):
    task_with_item_vars = []

    def group_task_transitions(self, mistral_transition_list):
        expr_transitions = ruamel.yaml.comments.CommentedMap()

        # If the transition is just a string, convert it to a list containing
        # the string
        if isinstance(mistral_transition_list, six.string_types):
            mistral_transition_list = [mistral_transition_list]

        simple_transitions = []
        for transition in mistral_transition_list:
            # if this is a string, then the transition is simply the name of the
            # next task
            if isinstance(transition, six.string_types):
                simple_transitions.append(transition)

            # if the transition is a dict, then it has an expression
            #  key = next task name
            #  value = conditional expression for when this transition should take place
            elif isinstance(transition, type_utils.dict_types):
                # group transitions by their common expression, this way we can keep
                # all of the transitions with the same expressions in the same
                # `when:` condition
                for task_name, expr in six.iteritems(transition):
                    if expr not in expr_transitions:
                        expr_transitions[expr] = []
                    expr_transitions[expr].append(task_name)
            else:
                raise TypeError('Task transition is not a "string" or "dict": {}  type={}'
                                .format(transition, type(transition)))

        return simple_transitions, expr_transitions

    def dict_to_list(self, d):
        return [{k: v} for k, v in six.iteritems(d)]

    def extract_context_variables(self, obj):
        # given an object that contains Orquesta YAQL or Jinja expressions (eg: contain 'ctx()'),
        # extract all variable names into a set
        variable_names = set()
        if isinstance(obj, type_utils.dict_types):
            for key in obj.keys():
                variable_names = variable_names | self.extract_context_variables(key)
            for val in obj.values():
                variable_names = variable_names | self.extract_context_variables(val)
        elif isinstance(obj, list):
            for element in obj:
                variable_names = variable_names | self.extract_context_variables(element)
        elif isinstance(obj, six.string_types):
            for mobj in CTX_RGX.finditer(obj):
                match = mobj.group(1) or mobj.group(2)
                variable_names.add(match)
        return variable_names

    def convert_task_transition_simple(self, transitions, publish, orquesta_expr, expr_converter):
        # if this is a simple name of a task:
        # on-success:
        #   - do_thing_a
        #   - do_thing_b
        #
        # this should produce the following in orquesta
        # next:
        #   - when: "{{ succeeded() }}"
        #     do:
        #       - do_thing_a
        #       - do_thing_b
        simple_transition = ruamel.yaml.comments.CommentedMap()

        # on-complete doesn't have an orquesta_expr, so we should not
        # add the 'when' clause in that case
        if orquesta_expr:
            simple_transition['when'] = expr_converter.wrap_expression(orquesta_expr)

        # add in published variables
        if publish:
            publish_converted = ExpressionConverter.convert_dict(publish)
            simple_transition['publish'] = self.dict_to_list(publish_converted)

        # add in the transition list
        if transitions:
            simple_transition['do'] = transitions

        return simple_transition

    def convert_task_transition_expr(self, task_name, expression_list, publish, orquesta_expr):
        # group all complex expressions by their common expression
        # this way we can keep all of the transitions with the same
        # expressions in the same `when:` condition
        #
        # on-success:
        #   - do_thing_a: "{{ _.x }}"
        #   - do_thing_b: "{{ _.x }}"
        #   - do_thing_c: "{{ not _.x }}"
        #
        # should produce the following in orquesta
        #
        # next:
        #   - when: "{{ succeeded() and _.x }}"
        #     do:
        #       - do_thing_a
        #       - do_thing_b
        #   - when: "{{ succeeded() and not _.x }}"
        #     do:
        #       - do_thing_c
        transitions = []
        for expr, task_list in six.iteritems(expression_list):
            expr_transition = ruamel.yaml.comments.CommentedMap()
            expr_converted = ExpressionConverter.convert(expr)

            # for some transitions (on-complete) the orquesta_expr may be empty
            # so only add it in, if it's necessary
            if orquesta_expr:
                converter = ExpressionConverter.get_converter(expr_converted)
                expr_converted = converter.unwrap_expression(expr_converted)
                o_expr = '{} and ({})'.format(orquesta_expr, expr_converted)
                o_expr = converter.wrap_expression(o_expr)
            else:
                o_expr = expr_converted

            expr_transition['when'] = o_expr
            if publish:
                converted_publish = ExpressionConverter.convert_dict(publish)
                expr_transition['publish'] = [{k: v} for k, v in converted_publish.items()]

                expr_transition['when'] = self.replace_immediately_referenced_variables(task_name,
                                                                                        expr_transition['when'],
                                                                                        converted_publish)

            expr_transition['do'] = task_list
            transitions.append(expr_transition)
        return transitions

    def replace_immediately_referenced_variables(self, task_name, when_expr, publish_dict):
        # Check for context variable references that are used in the 'when'
        # expression
        variables_in_when = self.extract_context_variables(when_expr)

        variables_in_publish = set(publish_dict.keys())

        immediately_referenced_variables = variables_in_when & variables_in_publish

        # Replace the context variable references in the 'when'
        # expression with their expression from the 'publish' block
        for variable in immediately_referenced_variables:
            # First off, make sure they are the same type of expression,
            # we don't want to inject a Jinja expression in the middle
            # of a YAQL expression
            when_expr_type = ExpressionConverter.expression_type(when_expr)
            publish_expr_type = ExpressionConverter.expression_type(publish_dict[variable])

            if when_expr_type == publish_expr_type:
                # Grab the variable expression
                converter = ExpressionConverter.get_converter(publish_dict[variable])
                unwrapped_expr = converter.unwrap_expression(publish_dict[variable])

                # Replace double parentheses
                # ((ctx().variable))    ->  (result().result['variable'] + 1)
                variable_reference_dp = r'\(\(ctx\(\).{var}\)\)'.format(var=variable)
                when_expr = re.sub(variable_reference_dp, "({})".format(unwrapped_expr), when_expr)

                # Don't add in parentheses if they already exist
                # (ctx().variable)      ->  (result().result['variable'] + 1)
                variable_reference_dp = r'\(ctx\(\).{var}\)'.format(var=variable)
                when_expr = re.sub(variable_reference_dp, "({})".format(unwrapped_expr), when_expr)

                # Keep surrounding context and add surrounding parentheses
                # (ctx().variable ...   ->  ((result().result['variable'] + 1) ...
                # ... ctx().variable)   ->  ... (result().result['variable'] + 1))
                # ...ctx().variable...  ->  ...(result().result['variable'] + 1)...
                variable_reference_ep = r'(.)\bctx\(\).{var}\b(.)'.format(var=variable)
                when_expr = re.sub(variable_reference_ep, r"\1({})\2".format(unwrapped_expr), when_expr)

                # Keep double enclosing internal parentheses
                # (ctx(variable))  ->  (result().result['variable'] + 1)
                variable_reference_eip = r'\(ctx\({var}\)\)'.format(var=variable)
                when_expr = re.sub(variable_reference_eip, "({})".format(unwrapped_expr), when_expr)

            else:
                warnings.warn("The transition \"{when_expr}\" in {task_name} "
                              "references the '{variable}' context variable, "
                              "which is published in the same transition. You "
                              "will need to manually convert the {variable} "
                              "expression in the transition."
                              .format(when_expr=when_expr,
                                      task_name=task_name,
                                      variable=variable)),
        return when_expr

    def default_task_transition_map(self):
        transitions = ruamel.yaml.comments.CommentedMap()
        transitions['on-success'] = ruamel.yaml.comments.CommentedMap()
        transitions['on-success']['publish'] = ruamel.yaml.comments.CommentedMap()
        transitions['on-success']['orquesta_expr'] = 'succeeded()'

        transitions['on-error'] = ruamel.yaml.comments.CommentedMap()
        transitions['on-error']['publish'] = ruamel.yaml.comments.CommentedMap()
        transitions['on-error']['orquesta_expr'] = 'failed()'

        transitions['on-complete'] = ruamel.yaml.comments.CommentedMap()
        transitions['on-complete']['publish'] = ruamel.yaml.comments.CommentedMap()
        transitions['on-complete']['orquesta_expr'] = None
        return transitions

    def normalize_transition_task_names(self, o_task_spec):
        for i, transition in enumerate(o_task_spec['next']):
            if 'do' in o_task_spec['next'][i]:
                new_dos = [task_utils.translate_task_name(tname) for tname in transition['do']]
                o_task_spec['next'][i]['do'] = new_dos
        return o_task_spec

    def convert_task_transitions(self, task_name, m_task_spec, expr_converter, wf_vars):
        # group all complex expressions by their common expression
        # this way we can keep all of the transitions with the same
        # expressions in the same `when:` condition
        #
        # publish:
        #   good_data: "{{ _.good }}"
        # publish-on-error:
        #   bad_data: "{{ _.bad }}"
        # on-success:
        #   - do_thing_a: "{{ _.x }}"
        #   - do_thing_b: "{{ _.x }}"
        # on-error:
        #   - do_thing_error: "{{ _.e }}"
        # on-complete:
        #   - do_thing_sometimes: "{{ _.d }}"
        #   - do_thing_always
        #
        # should produce the following in orquesta
        #
        # next:
        #   - when: "{{ succeeded() }}"
        #     publish:
        #       - good_data: "{{ ctx().good }}"
        #   - when: "{{ failed() }}"
        #     publish:
        #       - bad_data: "{{ ctx().bad }}"
        #   - when: "{{ succeeded() and ctx().x }}"
        #     publish:
        #       - good_data: "{{ ctx().good }}"
        #     do:
        #       - do_thing_a
        #       - do_thing_b
        #   - when: "{{ failed() and ctx().e }}"
        #     publish:
        #       - bad_data: "{{ ctx().bad }}"
        #     do:
        #       - do_thing_error
        #   - when: "{{ ctx().d }}"
        #     publish:
        #       - good_data: "{{ ctx().good }}"
        #       - bad_data: "{{ ctx().bad }}"
        #     do:
        #       - do_thing_sometimes
        #   - publish:
        #       - good_data: "{{ ctx().good }}"
        #       - bad_data: "{{ ctx().bad }}"
        #     do:
        #       - do_thing_always
        o_task_spec = ruamel.yaml.comments.CommentedMap()
        o_task_spec['next'] = []

        transitions = self.default_task_transition_map()

        if m_task_spec.get('publish'):
            transitions['on-success']['publish'] = m_task_spec['publish']

        if m_task_spec.get('publish-on-error'):
            transitions['on-error']['publish'] = m_task_spec['publish-on-error']

        if m_task_spec.get('on-complete'):
            # Handling on-complete publishing is more complicated, because the
            # dictionaries from publish and publish-on-error need to be
            # combined, but cannot overlap.
            publish = m_task_spec.get('publish', ruamel.yaml.comments.CommentedMap())
            publish_on_error = m_task_spec.get('publish-on-error',
                                               ruamel.yaml.comments.CommentedMap())

            # Generate a list of common keys for publish and publish-on-error
            # dictionaries
            common_publish_keys = list(set(publish.keys()) & set(publish_on_error.keys()))

            # If common keys have the same value, it's fine if we merge the
            # dictionaries.
            common_publish_keys_different_values = [
                k for k in common_publish_keys if publish[k] != publish_on_error[k]
            ]

            # If there are any common keys that correspond to different values
            # the publish dictionary and publish-on-error dictionaries, we need
            # to bail.
            if common_publish_keys_different_values:
                error_message = ("Task '{}' contains one or more keys ({}) in both publish and "
                                 "publish-on-error dictionaries that have different values. "
                                 "Please either remove the common keys, or ensure that the values "
                                 "of any common keys are the same.").\
                    format(task_name, ', '.join(common_publish_keys_different_values))
                raise NotImplementedError(error_message)

            publish_on_complete = publish.copy()
            publish_on_complete.update(publish_on_error)
            transitions['on-complete']['publish'] = publish_on_complete

        for m_transition_name, data in six.iteritems(transitions):
            m_transitions = m_task_spec.get(m_transition_name, [])
            trans_simple, trans_expr = self.group_task_transitions(m_transitions)

            publish_to_workflow_context = bool(wf_vars & set(data.get('publish').keys()))

            # Create a transition for the simple task lists
            if trans_simple or publish_to_workflow_context:
                o_trans_simple = self.convert_task_transition_simple(trans_simple,
                                                                     data.get('publish'),
                                                                     data['orquesta_expr'],
                                                                     expr_converter)
                o_task_spec['next'].append(o_trans_simple)

            # Create multiple transitions, one for each unique expression
            o_trans_expr_list = self.convert_task_transition_expr(task_name,
                                                                  trans_expr,
                                                                  data.get('publish'),
                                                                  data['orquesta_expr'])
            o_task_spec['next'].extend(o_trans_expr_list)

        o_task_spec = self.normalize_transition_task_names(o_task_spec)

        return o_task_spec if o_task_spec['next'] else ruamel.yaml.comments.CommentedMap()

    def convert_action(self, m_action):
        if m_action in UNSUPPORTED_MISTRAL_ACTIONS:
            raise NotImplementedError(("Action '{}' is not supported in orquesta.").
                                      format(m_action))

        o_action = MISTRAL_ACTION_CONVERSION_TABLE.get(m_action, m_action)

        kwargs = {'item_vars': self.task_with_item_vars}
        return MixedExpressionConverter.convert_string(o_action, **kwargs)

    def _get_inverse_comparison_operator(self, operator):
        # Should probably push this into the converters themselves,
        # but this will suffice for now
        # Convert and invert *simple* YAQL comparisons, since most continue-on's will
        # likely be simple comparisons
        return COMPARISON_OPERATOR_INVERSES[operator]

    def convert_retry(self, m_retry, task_name):
        o_retry = ruamel.yaml.comments.CommentedMap()
        if m_retry.get('count'):
            o_retry['count'] = m_retry['count']
        if m_retry.get('delay'):
            o_retry['delay'] = m_retry['delay']

        # Check for both and bail early
        if m_retry.get('continue-on') and m_retry.get('break-on'):
            raise NotImplementedError("Cannot convert \"retry\" spec with both a \"break-on\" and "
                                      "\"continue-on\" attributes in the \"{task_name}\" task."
                                      .format(task_name=task_name))
        elif m_retry.get('continue-on') or m_retry.get('break-on'):
            expr = m_retry.get('continue-on') or m_retry.get('break-on')

            converter = ExpressionConverter.get_converter(expr)
            if converter:
                expr = converter.unwrap_expression(expr)
                expr = converter.convert_string(expr)

                comp_op_match = RETRY_BREAK_ON_RGX.match(expr)

                if comp_op_match:
                    inv_op = self._get_inverse_comparison_operator(comp_op_match.group('comp_op'))
                    expr = '{lhs} {op} {rhs}'.format(
                        lhs=comp_op_match.group('lhs'),
                        op=inv_op,
                        rhs=comp_op_match.group('rhs'))
                    # TODO: For YAQL we can also simply do 'not (expr)'

                expr = converter.wrap_expression(expr)
            o_retry['when'] = expr

        return o_retry

    def convert_input(self, input_):
        kwargs = {'item_vars': self.task_with_item_vars}
        return MixedExpressionConverter.convert_dict(input_, **kwargs)

    def convert_with_items_expr(self, expression, expr_converter):
        # Convert all with-items attributes
        #
        # with-items:
        #   - i in <% $.yaql_data %>
        #
        # with-items:
        #   - i in <% $.yaql_data %>
        #   - j in <% $.more_data %>
        #   - k in <% $.all_the_data %>
        #
        # with-items: i in <% $.yaql_data %>
        # concurrency: 2
        #
        # with-items: i in {{ _.jinja_data }}
        #
        # with-items: i in [0, 1, 2, 3]
        #
        # should produce the following in orquesta
        #
        # with:
        #   items: i in <% ctx().yaql_data %>
        #
        # with:
        #   items: i, j, k in <% zip($.yaql_data, $.more_data, $.all_the_data) %>
        #
        # with:
        #   items: i in <% $.yaql_data %>
        #   concurrency: 2
        #
        # with:
        #   items: i in {{ ctx().jinja_data }}
        #
        # with:
        #   items: i in <% [0, 1, 2, 3] %>
        converter = None
        var_list = []
        expr_list = []
        if isinstance(expression, list):
            expression_list = expression
        else:
            # Create a list with a single element
            expression_list = [expression]

        for expr_item in expression_list:
            m = WITH_ITEMS_EXPR_RGX.match(expr_item)
            if m:
                var = m.group('var')
                expr = m.group('expr')
                converter = ExpressionConverter.get_converter(expr)
                if converter:
                    expr = converter.unwrap_expression(expr)
                    expr = converter.convert_string(expr)
                var_list.append(var)
                expr_list.append(expr)
            else:
                raise NotImplementedError("Unrecognized with-items expression: '{}'".
                                          format(expr_item))

        # If we have a list of expressions, we need to join them with commas
        # and feed that into 'zip()'
        # If we only have one expression in the list, we don't need to join
        # them or use 'zip()', we just use it directly
        if len(expr_list) > 1:
            expr_list_string = 'zip({expr})'.format(expr=', '.join(expr_list))
        else:
            expr_list_string = expr_list[0]

        # Default to the global expression converter if we could not determine one
        converter = converter if converter else expr_converter

        # We need to save the list of expression variables for when we convert
        # item access to item() instead of ctx()
        self.task_with_item_vars.extend(var_list)

        return "{vars} in {wrapped_expr}".format(
            vars=', '.join(var_list),
            wrapped_expr=converter.wrap_expression(expr_list_string))

    def convert_with_items(self, m_task_spec, expr_converter):
        with_items = m_task_spec['with-items']

        with_attr = {
            'items': self.convert_with_items_expr(with_items, expr_converter),
        }

        if m_task_spec.get('concurrency'):
            concurrency_expr = m_task_spec['concurrency']

            # Only try to convert the concurrency expression if it's a str
            if isinstance(concurrency_expr, six.string_types):
                converter = ExpressionConverter.get_converter(concurrency_expr)
                if converter:
                    concurrency_expr = converter.unwrap_expression(concurrency_expr)
                    concurrency_expr = converter.convert_string(concurrency_expr)
                    concurrency_expr = converter.wrap_expression(concurrency_expr)
            with_attr['concurrency'] = concurrency_expr

        return with_attr

    def convert_tasks(self, mistral_wf_tasks, expr_converter, wf_vars, force=False):
        orquesta_wf_tasks = ruamel.yaml.comments.CommentedMap()
        for task_name, m_task_spec in six.iteritems(mistral_wf_tasks):
            # The variables in with-items expressions need to be accessed using
            # the `item(var)` syntax, not the usual `ctx().var` syntax, so we
            # use this list to keep track of which variables are within the
            # task context
            self.task_with_item_vars = []
            o_task_spec = ruamel.yaml.comments.CommentedMap()
            if force:
                for attr in TASK_UNSUPPORTED_ATTRIBUTES:
                    val = m_task_spec.get(attr)
                    if val:
                        o_task_spec[attr] = val
            else:
                for attr in TASK_UNSUPPORTED_ATTRIBUTES:
                    if attr in m_task_spec:
                        raise NotImplementedError(("Task '{}' contains an attribute '{}'"
                                                   " that is not supported in orquesta.").
                                                  format(task_name, attr))

            if m_task_spec.get('with-items'):
                o_task_spec['with'] = self.convert_with_items(m_task_spec, expr_converter)

            if m_task_spec.get('action'):
                o_task_spec['action'] = self.convert_action(m_task_spec['action'])

            if m_task_spec.get('retry'):
                o_task_spec['retry'] = self.convert_retry(m_task_spec['retry'], task_name)

            if m_task_spec.get('join'):
                o_task_spec['join'] = m_task_spec['join']

            if m_task_spec.get('input'):
                o_task_spec['input'] = self.convert_input(m_task_spec['input'])

            o_task_transitions = self.convert_task_transitions(
                task_name, m_task_spec, expr_converter, wf_vars)
            o_task_spec.update(o_task_transitions)

            orquesta_wf_tasks[task_utils.translate_task_name(task_name)] = o_task_spec

        return orquesta_wf_tasks

    def expr_type_converter(self, expr_type):
        if expr_type is None:
            expr_converter = JinjaExpressionConverter()
        elif isinstance(expr_type, six.string_types):
            expr_type = expr_type.lower()
            if expr_type == 'jinja':
                expr_converter = JinjaExpressionConverter()
            elif expr_type == 'yaql':
                expr_converter = YaqlExpressionConverter()
            else:
                raise TypeError("Unknown expression type: {}".format(expr_type))
        elif isinstance(expr_type, BaseExpressionConverter):
            expr_converter = expr_type
        else:
            raise TypeError("Unknown expression class type: {}".format(type(expr_type)))
        return expr_converter

    def convert(self, mistral_wf, expr_type=None, force=False):
        variables_used_in_output = set()
        expr_converter = self.expr_type_converter(expr_type)
        orquesta_wf = ruamel.yaml.comments.CommentedMap()
        orquesta_wf['version'] = '1.0'

        if force:
            for attr in WORKFLOW_UNSUPPORTED_ATTRIBUTES:
                val = mistral_wf.get(attr)
                if val:
                    orquesta_wf[attr] = val
        else:
            for attr in WORKFLOW_UNSUPPORTED_ATTRIBUTES:
                if attr in mistral_wf:
                    raise NotImplementedError(("Workflow contains an attribute '{}' that is not"
                                               " supported in orquesta.").format(attr))

        if mistral_wf.get('description'):
            orquesta_wf['description'] = mistral_wf['description']

        if mistral_wf.get('type'):
            if mistral_wf['type'] not in WORKFLOW_TYPES:
                raise NotImplementedError(("Workflows of type '{}' are NOT supported."
                                           " Only 'direct' workflows can be converted").
                                          format(mistral_wf['type']))

        if mistral_wf.get('input'):
            orquesta_wf['input'] = ExpressionConverter.convert_list(mistral_wf['input'])

        if mistral_wf.get('vars'):
            expression_vars = ExpressionConverter.convert_dict(mistral_wf['vars'])
            orquesta_wf['vars'] = self.dict_to_list(expression_vars)

        if mistral_wf.get('output'):
            output = ExpressionConverter.convert_dict(mistral_wf['output'])
            orquesta_wf['output'] = self.dict_to_list(output)

            variables_used_in_output = self.extract_context_variables(output)

        if mistral_wf.get('tasks'):
            o_tasks = self.convert_tasks(
                mistral_wf['tasks'],
                expr_converter,
                variables_used_in_output,
                force=force)
            if o_tasks:
                orquesta_wf['tasks'] = o_tasks

        return orquesta_wf



from tests.base_test_case import BaseTestCase

from orquestaconvert.utils import task_utils


class TestTaskUtils(BaseTestCase):
    __test__ = True

    def test_convert_dashes_to_underscores(self):
        self.assertEqual(task_utils.translate_task_name('foo-bar-baz'), 'foo_bar_baz')
        self.assertEqual(task_utils.translate_task_name('baz_foo'), 'baz_foo')



import re
from orquestaconvert.expressions.base import BaseExpressionConverter

# {{ xxx }} -> xxx
UNWRAP_REGEX = r"({{)(.*)(}})"
UNWRAP_PATTERN = re.compile(UNWRAP_REGEX)

# _. -> ctx().
CONTEXT_VARS_REGEX = r"\b(_\.([\w]+))"
CONTEXT_VARS_PATTERN = re.compile(CONTEXT_VARS_REGEX)


class JinjaExpressionConverter(BaseExpressionConverter):

    @classmethod
    def wrap_expression(cls, expr):
        return "{{ " + expr + " }}"

    @classmethod
    def unwrap_expression(cls, expr):
        return UNWRAP_PATTERN.sub(cls._replace_unwrap, expr)

    @classmethod
    def convert_context_vars(cls, expr, **kwargs):
        _replace_vars = cls._get_replace_vars(**kwargs)
        return CONTEXT_VARS_PATTERN.sub(_replace_vars, expr)



import re
import six
import warnings

import ruamel.yaml.comments

from orquestaconvert.expressions import ExpressionConverter
from orquestaconvert.expressions.base import BaseExpressionConverter
from orquestaconvert.utils import type_utils


# These regexes match Jinja and YAQL expressions, respectively. They are used
# to pick out expressions from strings that contain both types of expressions,
# which we call 'mixed' expressions, in attributes like task actions:
#
# tasks:
#   - do_thing:
#       with-item: target_host in {{ _.target_hosts }}
#       action: ping <% $.ping_flags %> {{ _.target_host }}
#
# Each expression needs to be converted to use the ctx() and item() accessors
# in Orquesta:
#
# tasks:
#   - do_thing:
#       action: ping <% ctx().ping_flags %> {{ item(target_host) }}
#
JINJA_EXPR_RGX = re.compile(r'(?P<expr>(?:{{)\s*.+?\s*(?:}}))')
YAQL_EXPR_RGX = re.compile(r'(?P<expr>(?:<%)\s*.+?\s*(?:%>))')


class MixedExpressionConverter(BaseExpressionConverter):
    '''
    Mixed expressions are strings that possibly contain both Jinja and YAQL
    expressions. This converter is used to convert all expressions, regardless
    of their type.
    '''

    @classmethod
    def convert(cls, expr, **kwargs):
        if (isinstance(expr, type_utils.dict_types)):
            return cls.convert_dict(expr, **kwargs)
        elif isinstance(expr, list):
            return cls.convert_list(expr, **kwargs)
        elif isinstance(expr, six.string_types):
            return cls.convert_string(expr, **kwargs)
        elif isinstance(expr, bool):
            return expr
        elif expr is None:
            # Note: The only point to explicitly converting None is to skip
            #       emitting the warning. Otherwise we would remove this elif
            #       and let the function return the input. Also, ruamel doesn't
            #       explicitly serialize null to None, it just leaves the key
            #       blank (in other words, it "implicitly" serializes it).
            #       Example:
            #
            #           next:
            #             - when: ...
            #               publish:
            #                 - continue:
            #               ...
            #
            #       See this link for more information:
            #       https://bitbucket.org/ruamel/yaml/issues/169/roundtripdumper-dumps-null-values
            return None
        elif isinstance(expr, int):
            return expr
        else:
            warnings.warn("Could not recognize expression '{}'; results may not "
                          "be accurate.".format(expr), SyntaxWarning)
            return expr

    @classmethod
    def convert_string_containing_expressions(cls, match, **kwargs):
        expr = match.group('expr')
        converter = ExpressionConverter.get_converter(expr)
        if converter:
            expr = converter.unwrap_expression(expr)
            expr = converter.convert_string(expr, **kwargs)
            expr = converter.wrap_expression(expr)

        return expr

    @classmethod
    def convert_string(cls, expr, **kwargs):
        def _inner_convert_string(match):
            return cls.convert_string_containing_expressions(match, **kwargs)
        expr = JINJA_EXPR_RGX.sub(_inner_convert_string, expr)
        expr = YAQL_EXPR_RGX.sub(_inner_convert_string, expr)
        return expr

    @classmethod
    def convert_dict(cls, expr_dict, **kwargs):
        converted = ruamel.yaml.comments.CommentedMap()
        for k, expr in six.iteritems(expr_dict):
            converted[k] = cls.convert(expr, **kwargs)
        return converted

    @classmethod
    def convert_list(cls, expr_list, **kwargs):
        return [cls.convert(expr, **kwargs) for expr in expr_list]



from tests.base_test_case import BaseTestCase

from orquestaconvert.utils import type_utils

import ruamel.yaml


class TestTypeUtils(BaseTestCase):
    __test__ = True

    def test_dict_types(self):
        self.assertIsInstance({}, type_utils.dict_types)
        self.assertIsInstance(ruamel.yaml.comments.CommentedMap(), type_utils.dict_types)



from tests.base_test_case import BaseTestCase

from orquestaconvert.expressions.base import (
    AbstractBaseExpressionConverter,
    BaseExpressionConverter,
)


class AbstractBaseExpressionsTestCase(BaseTestCase):
    __test__ = True

    def test_wrap_expression_raises(self):
        with self.assertRaises(NotImplementedError):
            AbstractBaseExpressionConverter.wrap_expression('junk')

    def test_unwrap_expression_raises(self):
        with self.assertRaises(NotImplementedError):
            AbstractBaseExpressionConverter.unwrap_expression('junk')

    def test_convert_string_raises(self):
        with self.assertRaises(NotImplementedError):
            AbstractBaseExpressionConverter.convert_string('junk')

    def test_convert_context_vars_raises(self):
        with self.assertRaises(NotImplementedError):
            AbstractBaseExpressionConverter.convert_context_vars('junk')

    def test_convert_task_result_raises(self):
        with self.assertRaises(NotImplementedError):
            AbstractBaseExpressionConverter.convert_task_result('junk')

    def test_convert_st2kv(self):
        with self.assertRaises(NotImplementedError):
            AbstractBaseExpressionConverter.convert_st2kv('junk')

    def test_convert_st2_execution_id(self):
        with self.assertRaises(NotImplementedError):
            AbstractBaseExpressionConverter.convert_st2_execution_id('junk')

    def test_convert_st2_api_url(self):
        with self.assertRaises(NotImplementedError):
            AbstractBaseExpressionConverter.convert_st2_api_url('junk')


class BaseExpressionsTestCase(BaseTestCase):
    __test__ = True

    def test_wrap_expression_raises(self):
        with self.assertRaises(NotImplementedError):
            BaseExpressionConverter.wrap_expression('junk')

    def test_unwrap_expression_raises(self):
        with self.assertRaises(NotImplementedError):
            BaseExpressionConverter.unwrap_expression('junk')

    def test_convert_static_string(self):
        with self.assertRaises(NotImplementedError):
            BaseExpressionConverter.convert_string('junk')

    def test_convert_static_context_vars_raises(self):
        with self.assertRaises(NotImplementedError):
            BaseExpressionConverter.convert_context_vars('junk')



#!/usr/bin/env python

import argparse
import sys

from orquesta.specs.mistral.v2 import workflows as mistral_workflow
from orquesta.specs.native.v1 import models as orquesta_workflow

from orquestaconvert.workflows.base import WorkflowConverter
from orquestaconvert.utils import yaml_utils


class Client(object):

    def parser(self):
        parser = argparse.ArgumentParser(description='Convert Mistral workflows to Orquesta')
        parser.add_argument('-v', '--verbose', default=False, action='store_true',
                            help='Print success message when validating, otherwise ignored')
        parser.add_argument('-e', '--expressions',
                            choices=['jinja', 'yaql'],
                            default='jinja',
                            help=('Type of expressions to use when inserting new expressions'
                                  ' (things like task transitions in the "when" clause.'))
        parser.add_argument('--force', default=False, action='store_true',
                            help='Include unsupported attributes in the generated outputs')
        parser.add_argument('--validate', default=False, action='store_true',
                            help='Validate the Orquesta workflow')
        parser.add_argument('filename', metavar='FILENAME', nargs=1,
                            help='Path to the Mistral Workflow YAML file to convert')
        return parser

    def validate_workflow_spec(self, wf_spec):
        result = wf_spec.inspect_syntax()
        if result:
            raise ValueError(result)

    def convert_file(self, filename, expr_type=None):
        # parse the Mistral workflow from file
        mistral_wf_data, mistral_wf_data_ruamel = yaml_utils.read_yaml(filename)

        # validate the Mistral workflow before we start
        mistral_wf_spec = mistral_workflow.instantiate(mistral_wf_data)
        self.validate_workflow_spec(mistral_wf_spec)

        # convert Mistral -> Orquesta
        mistral_wf = mistral_wf_data_ruamel[mistral_wf_spec.name]
        workflow_converter = WorkflowConverter()
        orquesta_wf_data_ruamel = workflow_converter.convert(mistral_wf, expr_type,
                                                             force=self.args.force)
        orquesta_wf_data_str = yaml_utils.obj_to_yaml(orquesta_wf_data_ruamel)
        orquesta_wf_data = yaml_utils.yaml_to_obj(orquesta_wf_data_str)

        # validate we've generated a proper Orquesta workflow
        orquesta_wf_spec = orquesta_workflow.instantiate(orquesta_wf_data)
        if not self.args.force:
            self.validate_workflow_spec(orquesta_wf_spec)

        # write out the new Orquesta workflow to a YAML string
        return yaml_utils.obj_to_yaml(orquesta_wf_data_ruamel)

    def validate_file(self, filename):
        # parse the Orquesta workflow from file
        orquesta_wf_data, orquesta_wf_data_ruamel = yaml_utils.read_yaml(filename)

        # validate the Orquesta workflow
        orquesta_wf_spec = orquesta_workflow.instantiate(orquesta_wf_data)
        self.validate_workflow_spec(orquesta_wf_spec)

        if self.args.verbose:
            print("Successfully validated workflow from {}".format(filename))

    def run(self, argv, output_stream):
        # Write the file to the output_stream
        self.args = self.parser().parse_args(argv)
        expr_type = self.args.expressions
        if self.args.validate:
            for f in self.args.filename:
                self.validate_file(f)
        else:
            for f in self.args.filename:
                output_stream.write(self.convert_file(f, expr_type))
        return 0


if __name__ == '__main__':
    # Write the converted workflow to stdout
    sys.exit(Client().run(sys.argv[1:], sys.stdout))



from __future__ import print_function

from tests.base_test_case import BasePackClientRunTestCase

from orquestaconvert.pack_client import PackClient

import mock


class PackClientTestCase(BasePackClientRunTestCase):
    __test__ = True

    def setUp(self):
        super(PackClientTestCase, self).setUp()

        self.client = mock.MagicMock()
        self.client.run = mock.MagicMock(return_value=0)
        self.pack_client = PackClient()

    def test_get_mistral_workflow_files_in_p_dir(self):
        workflows = self.pack_client.get_workflow_files('mistral-v2', self.p_actions_dir)
        self.assertEqual(workflows, self.pristine_action_wfs)

    def test_get_orquesta_workflow_files_in_p_dir(self):
        workflows = self.pack_client.get_workflow_files('orquesta', self.p_actions_dir)
        self.assertEqual(workflows, {})

    def test_get_orquesta_workflow_files_in_o_dir(self):
        workflows = self.pack_client.get_workflow_files('orquesta', self.o_actions_dir)
        self.assertEqual(self.orquesta_action_wfs, workflows)

    def test_validate_nothing(self):
        args = ['--validate', '--actions-dir={}'.format(self.m_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(result, 0)

        self.assertEqual(self.client.run.call_count, 0)

    def test_validate_orquesta(self):
        args = ['--validate', '--actions-dir={}'.format(self.o_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(result, 0)

        self.assertEqual(self.client.run.call_count, len(self.action_passing_files))

        calls = [
            mock.call(['--validate', wf], self.stdout)
            for wf in self.orquesta_action_wfs.values()
        ]
        self.client.run.assert_has_calls(calls, any_order=True)

    def test_convert_pack(self):
        args = ['-e', 'yaql', '--actions-dir={}'.format(self.m_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(result, 0)

        self.assertEqual(self.client.run.call_count, len(self.action_files))

        expected = [
            ['-e', 'yaql', wf]
            for wf in self.action_wfs.values()
        ]
        self.assertItemsEqual([call_args[0][0] for call_args in self.client.run.call_args_list],
                              expected)



import six
import warnings

from tests.base_test_case import BaseTestCase

from orquestaconvert.client import Client


class TestEndToEnd(BaseTestCase):
    __test__ = True

    def setUp(self):
        super(TestEndToEnd, self).setUp()
        self.maxDiff = 20000
        self.client = Client()
        parser = self.client.parser()
        # Ensure that the client has an args attribute
        self.client.args = parser.parse_args(['file.yaml'])

    def e2e_from_file(self, m_wf_filename, o_wf_filename=None):
        if o_wf_filename is None:
            o_wf_filename = m_wf_filename
        fixture_path = self.get_fixture_path('mistral/' + m_wf_filename)
        result = self.client.convert_file(fixture_path)
        expected = self.get_fixture_content('orquesta/' + o_wf_filename)
        self.assertMultiLineEqual(result, expected)

    def test_e2e_force_option(self):
        parser = self.client.parser()
        self.client.args = parser.parse_args(['--force', 'file.yaml'])
        self.e2e_from_file('unsupported_attributes.yaml',
                           '../broken/unsupported_attributes.yaml')

    def test_e2e_nasa_apod_twitter_post(self):
        self.e2e_from_file('nasa_apod_twitter_post.yaml')

    def test_e2e_emptywee_test(self):
        with self.assertRaises(NotImplementedError):
            self.e2e_from_file('emptywee_test.yaml')

    def test_e2e_output_test(self):
        self.e2e_from_file('output_test.yaml')

    def test_e2e_transition_strings_test(self):
        self.e2e_from_file('transition_strings.yaml')

    def test_e2e_boolean_publish_parameters(self):
        self.e2e_from_file('boolean_publish_parameters.yaml')

    def test_e2e_null_publish_parameters(self):
        self.e2e_from_file('null_publish_parameters.yaml')

    def test_e2e_int_publish_parameters(self):
        self.e2e_from_file('int_publish_parameters.yaml')

    def test_e2e_convert_dashes_in_task_names_to_underscores(self):
        self.e2e_from_file('dashes_in_task_names.yaml')

    def test_e2e_immediately_referenced_context_variables(self):
        with warnings.catch_warnings(record=True) as ws:
            self.e2e_from_file('immediately_referenced_context_variables.yaml')

            self.assertGreater(len(ws), 0)
            expected_warning = (
                "The transition \"{{ succeeded() and (ctx().operations) }}\" "
                "in task_1 references the 'operations' context variable, which "
                "is published in the same transition. You will need to "
                "manually convert the operations expression in the transition.")
            if six.PY2:
                self.assertEqual(expected_warning, ws[0].message.message)
            else:
                self.assertEqual(expected_warning, str(ws[0].message))



#!/usr/bin/env python

import argparse
import glob
import os
import shutil
import six
import sys

import yaml

from orquestaconvert.client import Client
from orquestaconvert.utils import yaml_utils


# The file extension to use for backup files - these files are backups of the
# original workflow and action metadata files, and normally be removed upon
# success or failure. However, sending a SIGINT or SIGHUP to the process may
# leave these files behind.
BACKUP_EXTENSION = 'orquestaconvert.bak.yaml'
# The file extension to use for temporary files. These files are the converted
# workflow and action metadata files, and should also be removed upon success
# or failure. However, it may sometimes be useful to keep these files,
# especially when using the --force option.
TMP_EXTENSION = 'orquesta.temp.yaml'


class PackClient(object):
    def parser(self):
        parser = argparse.ArgumentParser(description='Convert all Mistral workflows in a pack')
        parser.add_argument('--validate', default=False, action='store_true',
                            help='Validate the Orquesta workflow')
        parser.add_argument('--list-workflows', dest='workflow_type', type=str,
                            help='List Mistral workflows in the pack and exit')
        parser.add_argument('--actions-dir', dest='actions_directory', default=None, type=str,
                            help='The action directory to convert')
        return parser

    def get_workflow_files(self, workflow_type, directory=None):
        action_directory = directory if directory else 'actions'
        glob_string = '{}/*.yaml'.format(action_directory)
        a_files = glob.glob(glob_string)
        mistral_workflows = {}
        for a_file in a_files:
            with open(a_file, 'r') as f:
                action_data = yaml.safe_load(f.read())
            runner = action_data.get('runner_type')

            if runner == workflow_type:
                mistral_workflows[a_file] = os.path.join(
                    action_directory,
                    *os.path.split(action_data.get('entry_point')))

        return mistral_workflows

    def run(self, argv, output_stream, client=None):
        self.args, args = self.parser().parse_known_args(argv)
        wf_type = self.args.workflow_type
        directory = self.args.actions_directory
        if wf_type:
            for action, workflow in self.get_workflow_files(wf_type, directory).items():
                output_stream.write("{} --> {}\n".format(action, workflow))
            return 0

        if self.args.validate:
            args.append('--validate')
            filenames = self.get_workflow_files('orquesta', directory).values()
            total = 0
            for f in filenames:
                fargs = list(args)
                fargs.append(f)
                total += client.run(fargs, output_stream)
            return total

        filenames = self.get_workflow_files('mistral-v2', directory)

        exceptions = {}
        for a_f, m_f in six.iteritems(filenames):
            fargs = list(args)  # make a copy
            fargs.append(m_f)

            # Get the backup filenames
            m_f_backup = '{}.{}'.format(m_f, BACKUP_EXTENSION)
            a_f_backup = '{}.{}'.format(a_f, BACKUP_EXTENSION)
            o_f = '{}.{}'.format(m_f, TMP_EXTENSION)  # Orquesta workflow file

            # Convert the workflow and save the YAML string in output
            output = six.moves.StringIO()
            # In this next block of code we are attempting to create a filesystem
            # transaction, where we either want to commit all of the changes to
            # disk or none of the changes.
            # Usually you want to minimize what you put into a try block, so you
            # can catch different errors and handle them differently. However, in
            # this case, there are so many file operations that can fail that it
            # ends up nesting try/except/else blocks very deep, and exception
            # handlers simply do cleanup and re-raise the exception.
            # So instead, we do all of our work in the try block, and handle
            # cleaning up after the different failure conditions in the except
            # block, and handle success conditions in the else block.
            try:
                client.run(fargs, output)

                # Write the workflow file
                with open(o_f, 'w') as o_file:
                    o_file.write(str(output.getvalue()))

                # If the backup files already exist, they were created by a
                # previous run. In that case, we want to preserve the original
                # backup file, because it is more likely a valid Mistral
                # workflow than whatever the current workflow file is.
                if not os.path.isfile(m_f_backup):
                    # Move the existing workflow file to a backup
                    os.rename(m_f, m_f_backup)

                if not os.path.isfile(a_f_backup):
                    # Move the existing metadata file to a backup
                    shutil.copy2(a_f, a_f_backup)

                # Read the file into a dict, tweak the runner_type, and write
                # the dict back into a string
                action_data, action_data_ruamel = yaml_utils.read_yaml(a_f)
                action_data_ruamel['runner_type'] = 'orquesta'
                write_data = yaml_utils.obj_to_yaml(action_data_ruamel)

                # Promote/move the new workflow file into place
                shutil.move(o_f, m_f)

                # Rewrite the metadata file with the tweaked runner_rtype
                with open(a_f, 'w') as a_file:
                    a_file.write(write_data)
                # SUCCESS!

            except Exception as e:
                # Anything could have happened, so we check for bad conditions

                # If we have a backup workflow file, revert it
                if os.path.isfile(m_f_backup):
                    # Remove the converted workflow file
                    if os.path.isfile(m_f):
                        os.remove(m_f)
                    # Move the backup file back
                    os.rename(m_f_backup, m_f)

                # If we have a backup action metadata file
                if os.path.isfile(a_f_backup):
                    # Remove the converted metadata file
                    if os.path.isfile(a_f):
                        os.remove(a_f)
                    # Move the backup file back
                    os.rename(a_f_backup, a_f)

                # If we ever support just Python 3, we can add the exception
                # directly to the dictionary value:
                #
                # .append({'file': m_f, 'exception': e})
                #
                exceptions.setdefault(str(e), []).append(m_f)
            else:
                # Remove the backup files
                os.remove(m_f_backup)
                os.remove(a_f_backup)

        if exceptions:
            sys.stderr.write("ERROR: Unable to convert all Mistral workflows.\n")
            for e, wfs in exceptions.items():
                sys.stderr.write("ISSUE: {}\n".format(e))
                sys.stderr.write("Affected files:\n")
                for wf in wfs:
                    sys.stderr.write("  - {}\n".format(wf))
                sys.stderr.write("\n")

        return len(exceptions)


if __name__ == '__main__':
    sys.exit(PackClient().run(sys.argv[1:], sys.stdout, client=Client()))



from tests.base_test_case import BaseCLITestCase

from orquestaconvert.client import Client

import mock


class TestClient(BaseCLITestCase):
    __test__ = True

    def setUp(self):
        super(TestClient, self).setUp()
        self.client = Client()
        self.maxDiff = 20000

    def _validate_args(self, args, filename, expected_filename=None):
        # path to fixture file
        fixture_path = self.get_fixture_path('mistral/' + filename)

        # run
        exit_status = Client().run(args + [fixture_path], self.stdout)
        self.assertEqual(exit_status, 0)

        # read expected data
        if not expected_filename:
            expected_filename = filename
        expected = self.get_fixture_content('orquesta/' + expected_filename)

        # compare
        self.stdout.seek(0)
        stdout = self.stdout.read()
        self.assertMultiLineEqual(stdout, expected)

    def test_run(self):
        self._validate_args([], 'nasa_apod_twitter_post.yaml')

    def test_run_jinja(self):
        self._validate_args(['-e', 'jinja'], 'nasa_apod_twitter_post.yaml')

    def test_run_yaql(self):
        self._validate_args(['-e', 'yaql'],
                            'nasa_apod_twitter_post_yaql.yaml',
                            'nasa_apod_twitter_post_yaql.yaml')

    def test_run_expression_jinja(self):
        self._validate_args(['--expression', 'jinja'], 'nasa_apod_twitter_post.yaml')

    def test_run_expression_yaql(self):
        self._validate_args(['--expression', 'yaql'],
                            'nasa_apod_twitter_post_yaql.yaml',
                            'nasa_apod_twitter_post_yaql.yaml')

    def test_validate_workflow_spec_raises(self):
        wf_spec = mock.MagicMock()
        wf_spec.inspect_syntax.return_value = "some error string"

        with self.assertRaises(ValueError):
            self.client.validate_workflow_spec(wf_spec)



import abc
import re
import six

# task('xxx').result -> result()
#    if 'xxx' != current task name, error
TASK_RESULT_REGEX = r"(task\([\"\']?\w+[\"\']?\).result)"
TASK_RESULT_PATTERN = re.compile(TASK_RESULT_REGEX)

# - st2kv. -> st2kv('xxx')
ST2KV_REGEX = r"(st2kv\.([\w\.]+))"
ST2KV_PATTERN = re.compile(ST2KV_REGEX)

# env().st2_execution_id -> ctx().st2.action_execution_id
ST2_EXECUTION_ID_REGEX = r"\benv\(\)\.st2_execution_id\b"
ST2_EXECUTION_ID_PATTERN = re.compile(ST2_EXECUTION_ID_REGEX)

# env().st2_action_api_url -> ctx().st2.api_url
ST2_API_URL_REGEX = r"\benv\(\).st2_action_api_url\b"
ST2_API_URL_PATTERN = re.compile(ST2_API_URL_REGEX)


@six.add_metaclass(abc.ABCMeta)
class AbstractBaseExpressionConverter(object):

    @classmethod
    @abc.abstractmethod
    def wrap_expression(cls, expr):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def unwrap_expression(cls, expr):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def convert_string(self, expr, **kwargs):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def convert_context_vars(self, expr, **kwargs):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def convert_task_result(self, expr, **kwargs):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def convert_st2kv(self, expr, **kwargs):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def convert_st2_execution_id(self, expr, **kwargs):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def convert_st2_api_url(self, expr, **kwargs):
        raise NotImplementedError()


class BaseExpressionConverter(AbstractBaseExpressionConverter):

    @classmethod
    def _replace_unwrap(cls, match):
        return match.group(2).strip()

    @classmethod
    def convert_string(cls, expr, **kwargs):
        # others?
        expr = cls.convert_context_vars(expr, **kwargs)
        expr = cls.convert_task_result(expr)
        expr = cls.convert_st2kv(expr)
        expr = cls.convert_st2_execution_id(expr)
        expr = cls.convert_st2_api_url(expr)
        return expr

    @classmethod
    def _replace_item_vars(cls, match):
        return "item(" + match.group(2) + ")"

    @classmethod
    def _replace_context_vars(cls, match):
        return "ctx()." + match.group(2)

    @classmethod
    def _get_replace_vars(cls, **kwargs):
        item_vars = kwargs.get('item_vars') or []

        def _inner_replace_vars(match):
            if match and match.group(2) in item_vars:
                return cls._replace_item_vars(match)
            else:
                return cls._replace_context_vars(match)

        return _inner_replace_vars

    @classmethod
    def _replace_task_result(cls, match):
        return "result()"

    @classmethod
    def convert_task_result(cls, expr):
        # TODO error if task name is not the same task in this context
        return TASK_RESULT_PATTERN.sub(cls._replace_task_result, expr)

    @classmethod
    def _replace_st2kv(cls, match):
        return "st2kv('" + match.group(2) + "')"

    @classmethod
    def convert_st2kv(cls, expr):
        return ST2KV_PATTERN.sub(cls._replace_st2kv, expr)

    @classmethod
    def _replace_st2_execution_id(cls, match):
        return "ctx().st2.action_execution_id"

    @classmethod
    def convert_st2_execution_id(cls, expr):
        return ST2_EXECUTION_ID_PATTERN.sub(cls._replace_st2_execution_id, expr)

    @classmethod
    def _replace_st2_api_url(cls, match):
        return "ctx().st2.api_url"

    @classmethod
    def convert_st2_api_url(cls, expr):
        return ST2_API_URL_PATTERN.sub(cls._replace_st2_api_url, expr)



import re

from tests.base_test_case import BaseTestCase

from orquestaconvert.expressions.jinja import JinjaExpressionConverter
from orquestaconvert.expressions.yaql import YaqlExpressionConverter
from orquestaconvert.workflows.base import WorkflowConverter, COMPARISON_OPERATOR_INVERSES

import ruamel.yaml
OrderedMap = ruamel.yaml.comments.CommentedMap


BaseTestCase.maxDiff = None


class TestWorkflows(BaseTestCase):
    __test__ = True

    def __init__(self, *args, **kwargs):
        super(TestWorkflows, self).__init__(*args, **kwargs)
        self.maxDiff = None

    def test_group_task_transitions(self):
        converter = WorkflowConverter()
        transitions_list = [
            "simple transition string",
            {"key": "expr"},
            "another simple transition string",
            {"key2": "expression"},
            {"key3": "expr"},
        ]
        simple, expr = converter.group_task_transitions(transitions_list)
        self.assertEqual(simple, ["simple transition string",
                                  "another simple transition string"])
        expected = OrderedMap([("expr", ["key", "key3"]),
                               ("expression", ["key2"])])
        self.assertEqual(expr, expected)

    def test_group_task_string_transition(self):
        converter = WorkflowConverter()
        transitions_string = 'next_task'
        simple, expr = converter.group_task_transitions(transitions_string)
        self.assertEqual(simple, ['next_task'])

    def test_group_task_transitions_raises_bad_type(self):
        converter = WorkflowConverter()
        transitions_list = [["list is bad"]]
        with self.assertRaises(TypeError):
            converter.group_task_transitions(transitions_list)

    def test_dict_to_list(self):
        converter = WorkflowConverter()
        d = OrderedMap([('key1', 'value1'),
                        ('key2', 'value2'),
                        ('key3', 'value3')])
        result = converter.dict_to_list(d)
        self.assertEqual(result, [{'key1': 'value1'},
                                  {'key2': 'value2'},
                                  {'key3': 'value3'}])

    def test_extract_context_variables(self):
        converter = WorkflowConverter()
        output_block = OrderedMap([
            ('key1', '{{ ctx().value1_str }}'),
            ('list_<%% ctx().list_key2 %>', [
                'a',
                '<%% ctx().value2_list %>',
                {
                    'inner_key_1': 'inner_value_1',
                    'inner_key_2': '<%% ctx().inner_value_2 %>',
                    'inner_<%% ctx().inner_key_3 %>': 'inner_value_3',
                    'inner_{{ ctx().inner_key_4 }}': '<%% ctx().list_key2 %>',
                },
            ]),
            ('key3_int', 1),
            ('key4_bool_false', False),
            ('key5_bool_true', True),
            ('key6_null', None),
            ('key7_float', 1.0),
        ])
        expected_output = set([
            'inner_key_3',
            'inner_key_4',
            'inner_value_2',
            'list_key2',
            'value1_str',
            'value2_list',
        ])
        actual_output = converter.extract_context_variables(output_block)
        self.assertEqual(actual_output, expected_output)

    def test_convert_task_transition_simple(self):
        converter = WorkflowConverter()
        transitions = ['a', 'b', 'c']
        publish = OrderedMap([('key_plain', 'data'),
                              ('key_expr', '{{ _.test }}')])
        orquesta_expr = 'succeeded()'
        expr_converter = JinjaExpressionConverter()

        result = converter.convert_task_transition_simple(transitions,
                                                          publish,
                                                          orquesta_expr,
                                                          expr_converter)

        expected = OrderedMap([
            ('when', '{{ succeeded() }}'),
            ('publish', [
                {'key_plain': 'data'},
                {'key_expr': '{{ ctx().test }}'}
            ]),
            ('do', ['a', 'b', 'c']),
        ])
        self.assertEqual(result, expected)

    def test_convert_task_transition_simple_yaql(self):
        converter = WorkflowConverter()
        transitions = ['a', 'b', 'c']
        publish = OrderedMap([('key_plain', 'data'),
                              ('key_expr', '{{ _.test }}')])
        orquesta_expr = 'succeeded()'
        expr_converter = YaqlExpressionConverter()

        result = converter.convert_task_transition_simple(transitions,
                                                          publish,
                                                          orquesta_expr,
                                                          expr_converter)

        # note: only the orquesta_expr is converted using YAQL, the expressions
        #       in the rest are converted independently because they may each
        #       be a different type
        expected = OrderedMap([
            ('when', '<% succeeded() %>'),
            ('publish', [
                {'key_plain': 'data'},
                {'key_expr': '{{ ctx().test }}'}
            ]),
            ('do', ['a', 'b', 'c']),
        ])
        self.assertEqual(result, expected)

    def test_convert_task_transition_simple_no_orquesta_expr(self):
        converter = WorkflowConverter()
        transitions = ['a', 'b', 'c']
        publish = OrderedMap([('key_plain', 'data'),
                              ('key_expr', '{{ _.test }}')])
        orquesta_expr = None
        expr_converter = JinjaExpressionConverter()

        result = converter.convert_task_transition_simple(transitions,
                                                          publish,
                                                          orquesta_expr,
                                                          expr_converter)

        expected = OrderedMap([
            ('publish', [
                {'key_plain': 'data'},
                {'key_expr': '{{ ctx().test }}'}
            ]),
            ('do', ['a', 'b', 'c']),
        ])
        self.assertEqual(result, expected)

    def test_convert_task_transition_simple_no_publish(self):
        converter = WorkflowConverter()
        transitions = ['a', 'b', 'c']
        publish = None
        orquesta_expr = 'succeeded()'
        expr_converter = JinjaExpressionConverter()

        result = converter.convert_task_transition_simple(transitions,
                                                          publish,
                                                          orquesta_expr,
                                                          expr_converter)

        expected = OrderedMap([
            ('when', '{{ succeeded() }}'),
            ('do', ['a', 'b', 'c']),
        ])
        self.assertEqual(result, expected)

    def test_convert_task_transition_simple_no_transitions(self):
        converter = WorkflowConverter()
        transitions = None
        publish = OrderedMap([('key_plain', 'data'),
                              ('key_expr', '{{ _.test }}')])
        orquesta_expr = 'succeeded()'
        expr_converter = JinjaExpressionConverter()

        result = converter.convert_task_transition_simple(transitions,
                                                          publish,
                                                          orquesta_expr,
                                                          expr_converter)

        expected = OrderedMap([
            ('when', '{{ succeeded() }}'),
            ('publish', [
                {'key_plain': 'data'},
                {'key_expr': '{{ ctx().test }}'}
            ]),
        ])
        self.assertEqual(result, expected)

    def test_convert_task_transition_expr(self):
        converter = WorkflowConverter()
        expression_list = OrderedMap([('<% $.test %>', ['task1', 'task3']),
                                      ('{{ _.other }}', ['task2'])])
        orquesta_expr = 'succeeded()'

        result = converter.convert_task_transition_expr('task_name',
                                                        expression_list,
                                                        {},
                                                        orquesta_expr)

        expected = [
            OrderedMap([
                ('when', '<% succeeded() and (ctx().test) %>'),
                ('do', ['task1', 'task3']),
            ]),
            OrderedMap([
                ('when', '{{ succeeded() and (ctx().other) }}'),
                ('do', ['task2']),
            ]),
        ]
        self.assertEqual(result, expected)

    def test_convert_task_transition_expr_no_orquesta_expr(self):
        converter = WorkflowConverter()
        expression_list = OrderedMap([('<% $.test %>', ['task1', 'task3']),
                                      ('{{ _.other }}', ['task2'])])
        orquesta_expr = None

        result = converter.convert_task_transition_expr('task_name',
                                                        expression_list,
                                                        {},
                                                        orquesta_expr)

        expected = [
            OrderedMap([
                ('when', '<% ctx().test %>'),
                ('do', ['task1', 'task3']),
            ]),
            OrderedMap([
                ('when', '{{ ctx().other }}'),
                ('do', ['task2']),
            ]),
        ]
        self.assertEqual(result, expected)

    def test_default_task_transition_map(self):
        converter = WorkflowConverter()
        result = converter.default_task_transition_map()
        self.assertEqual(result, OrderedMap([
            ('on-success', OrderedMap([
                ('publish', OrderedMap()),
                ('orquesta_expr', 'succeeded()'),
            ])),
            ('on-error', OrderedMap([
                ('publish', OrderedMap()),
                ('orquesta_expr', 'failed()'),
            ])),
            ('on-complete', OrderedMap([
                ('publish', OrderedMap()),
                ('orquesta_expr', None),
            ])),
        ]))

    def test_convert_task_transitions(self):
        converter = WorkflowConverter()
        task_spec = OrderedMap([
            ('publish', OrderedMap([
                ('good_data', '{{ _.good }}'),
            ])),
            ('publish-on-error', OrderedMap([
                ('bad_data', '{{ _.bad }}'),
            ])),
            ('on-success', [
                OrderedMap([('do_thing_a', '{{ _.x }}')]),
                OrderedMap([('do_thing_b', '{{ _.x }}')])
            ]),
            ('on-error', [
                OrderedMap([('do_thing_error', '{{ _.e }}')]),
            ]),
            ('on-complete', [
                OrderedMap([('do_thing_sometimes', '{{ _.d }}')]),
                'do_thing_always'
            ]),
        ])
        expr_converter = JinjaExpressionConverter()
        result = converter.convert_task_transitions("task_name", task_spec, expr_converter, set())
        self.assertDictEqual(result, OrderedMap([
            ('next', [
                OrderedMap([
                    ('when', '{{ succeeded() and (ctx().x) }}'),
                    ('publish', [
                        {'good_data': '{{ ctx().good }}'}
                    ]),
                    ('do', [
                        'do_thing_a',
                        'do_thing_b',
                    ]),
                ]),
                OrderedMap([
                    ('when', '{{ failed() and (ctx().e) }}'),
                    ('publish', [
                        {'bad_data': '{{ ctx().bad }}'}
                    ]),
                    ('do', [
                        'do_thing_error',
                    ]),
                ]),
                OrderedMap([
                    ('publish', [
                        {'good_data': '{{ ctx().good }}'},
                        {'bad_data': '{{ ctx().bad }}'},
                    ]),
                    ('do', [
                        'do_thing_always',
                    ]),
                ]),
                OrderedMap([
                    ('when', '{{ ctx().d }}'),
                    ('publish', [
                        {'good_data': '{{ ctx().good }}'},
                        {'bad_data': '{{ ctx().bad }}'},
                    ]),
                    ('do', [
                        'do_thing_sometimes',
                    ]),
                ]),
            ]),
        ]))

    def test_convert_task_transitions_empty(self):
        converter = WorkflowConverter()
        task_spec = OrderedMap([])
        expr_converter = JinjaExpressionConverter()
        result = converter.convert_task_transitions("task_name", task_spec, expr_converter, set())
        self.assertEqual(result, OrderedMap([]))

    def test_convert_task_transitions_common_publish_keys_same_values(self):
        converter = WorkflowConverter()
        task_spec = OrderedMap([
            ('publish', OrderedMap([
                ('good_data', '{{ _.good }}'),
                ('common_data_1', '{{ _.common_1 }}'),
                ('common_data_2', '{{ _.common_2 }}'),
            ])),
            ('publish-on-error', OrderedMap([
                ('bad_data', '{{ _.bad }}'),
                ('common_data_1', '{{ _.common_1 }}'),
                ('common_data_2', '{{ _.common_2 }}'),
            ])),
            ('on-complete', [
                'do_thing_always'
            ]),
        ])
        expr_converter = JinjaExpressionConverter()

        result = converter.convert_task_transitions("task_name", task_spec, expr_converter, set())
        self.assertDictEqual(result, OrderedMap([
            ('next', [
                OrderedMap([
                    ('publish', [
                        {'good_data': '{{ ctx().good }}'},
                        {'common_data_1': '{{ ctx().common_1 }}'},
                        {'common_data_2': '{{ ctx().common_2 }}'},
                        {'bad_data': '{{ ctx().bad }}'},
                    ]),
                    ('do', [
                        'do_thing_always',
                    ]),
                ]),
            ]),
        ]))

    def test_convert_task_transitions_common_publish_keys_different_values(self):
        converter = WorkflowConverter()
        task_spec = OrderedMap([
            ('publish', OrderedMap([
                ('good_data', '{{ _.good }}'),
                ('common_key_1', '{{ _.common_data }}'),
                ('common_key_2', '{{ _.different_data_1_1 }}'),
                ('common_key_3', '{{ _.different_data_1_2 }}'),
            ])),
            ('publish-on-error', OrderedMap([
                ('bad_data', '{{ _.bad }}'),
                ('common_key_1', '{{ _.common_data }}'),
                ('common_key_2', '{{ _.different_data_2_1 }}'),
                ('common_key_3', '{{ _.different_data_2_2 }}'),
            ])),
            ('on-complete', [
                OrderedMap([('do_thing_sometimes', '{{ _.d }}')]),
                'do_thing_always'
            ]),
        ])
        expr_converter = JinjaExpressionConverter()

        with self.assertRaises(NotImplementedError) as ctx_m:
            converter.convert_task_transitions("tsk_name", task_spec, expr_converter, set())

        rgx = re.compile(r"Task 'tsk_name' contains one or more keys "
                         r"\((?:common_key_2, common_key_3|common_key_3, common_key_2)\) "
                         r"in both publish and publish-on-error dictionaries that have different "
                         r"values\. Please either remove the common keys, or ensure that the "
                         r"values of any common keys are the same\.")
        self.assertRegex(str(ctx_m.exception), rgx)

    def test_convert_with_items(self):
        wi = {
            'with-items': 'b in <% [3, 4, 5] %>',
        }
        converter = WorkflowConverter()
        actual = converter.convert_with_items(wi, YaqlExpressionConverter)
        # This should NOT have a concurrency key
        expected = {
            'items': 'b in <% [3, 4, 5] %>',
        }
        self.assertEqual(expected, actual)

    def test_convert_with_items_concurrency_integer(self):
        wi = {
            'with-items': 'b in <% [3, 4, 5] %>',
            'concurrency': 2,
        }
        converter = WorkflowConverter()
        actual = converter.convert_with_items(wi, YaqlExpressionConverter)
        # This must have a concurrency key
        expected = {
            'items': 'b in <% [3, 4, 5] %>',
            'concurrency': 2,
        }
        self.assertEqual(expected, actual)

    def test_convert_with_items_concurrency_string(self):
        wi = {
            'with-items': 'b in <% [3, 4, 5] %>',
            'concurrency': '2',
        }
        converter = WorkflowConverter()
        actual = converter.convert_with_items(wi, YaqlExpressionConverter)
        # This must have a concurrency key
        expected = {
            'items': 'b in <% [3, 4, 5] %>',
            'concurrency': '2',
        }
        self.assertEqual(expected, actual)

    def test_convert_with_items_concurrency_yaql(self):
        wi = {
            'with-items': 'b in <% [3, 4, 5] %>',
            'concurrency': '<% $.count %>',
        }
        converter = WorkflowConverter()
        actual = converter.convert_with_items(wi, YaqlExpressionConverter)
        # This must have a concurrency key
        expected = {
            'items': 'b in <% [3, 4, 5] %>',
            'concurrency': '<% ctx().count %>',
        }
        self.assertEqual(expected, actual)

    def test_convert_with_items_concurrency_jinja(self):
        wi = {
            'with-items': 'b in <% [3, 4, 5] %>',
            'concurrency': '{{ _.count }}',
        }
        converter = WorkflowConverter()
        actual = converter.convert_with_items(wi, YaqlExpressionConverter)
        # This must have a concurrency key
        expected = {
            'items': 'b in <% [3, 4, 5] %>',
            'concurrency': '{{ ctx().count }}',
        }
        self.assertEqual(expected, actual)

    def test_convert_with_items_expr_list(self):
        wi_list = [
            'a in <% [0, 1, 2] %>',
            'b in [3, 4, 5]',
            'c in <% $.all_the_things %>',
        ]
        converter = WorkflowConverter()
        actual = converter.convert_with_items_expr(wi_list, YaqlExpressionConverter)
        expected = "a, b, c in <% zip([0, 1, 2], [3, 4, 5], ctx().all_the_things) %>"
        self.assertEqual(expected, actual)

    def test_convert_with_items_expr_list_one_element(self):
        # Check that with-items expression lists with a single element don't
        # get put into a zip() expression
        wi_list = [
            'a in <% [0, 1, 2] %>',
        ]
        converter = WorkflowConverter()
        actual = converter.convert_with_items_expr(wi_list, YaqlExpressionConverter)
        expected = "a in <% [0, 1, 2] %>"
        self.assertEqual(expected, actual)

    def test_convert_with_items_expr_list_unrecognized_expression(self):
        wi_list = [
            'a in <% [0, 1, 2] %>',
            'BLARGETH',  # bad syntax
            'c in <% $.all_the_things %>',
        ]
        converter = WorkflowConverter()
        with self.assertRaises(NotImplementedError):
            converter.convert_with_items_expr(wi_list, YaqlExpressionConverter)

    def test_convert_with_items_expr_matching_regex_str(self):
        wi_str = 'b in <% [3, 4, 5] %>'
        converter = WorkflowConverter()
        actual = converter.convert_with_items_expr(wi_str, YaqlExpressionConverter)
        expected = 'b in <% [3, 4, 5] %>'
        self.assertEqual(expected, actual)

        wi_str = 'i in <% $.items %>'
        converter = WorkflowConverter()
        actual = converter.convert_with_items_expr(wi_str, YaqlExpressionConverter)
        expected = 'i in <% ctx().items %>'
        self.assertEqual(expected, actual)

    def test_convert_with_items_expr_nonmatching_regex_str(self):
        wi_str = 'b in [3, 4, 5]'
        converter = WorkflowConverter()
        actual = converter.convert_with_items_expr(wi_str, YaqlExpressionConverter)
        expected = 'b in <% [3, 4, 5] %>'
        self.assertEqual(expected, actual)

    def test_convert_with_items_expr_unrecognized_expression(self):
        wi_str = 'BLARGETH'
        converter = WorkflowConverter()
        with self.assertRaises(NotImplementedError):
            converter.convert_with_items_expr(wi_str, YaqlExpressionConverter)

    def test_retry(self):
        retry = {
            'count': 30,
            'delay': 5,
        }
        converter = WorkflowConverter()
        actual = converter.convert_retry(retry, 'retry_task')
        expected = OrderedMap([
            ('count', 30),
            ('delay', 5),
        ])
        self.assertEqual(expected, actual)

    def test_retry_continue_on(self):
        # unittest2's alternative to @pytest.mark.parametrize
        # https://docs.python.org/3/library/unittest.html#distinguishing-test-iterations-using-subtests  # noqa: E501
        for operator in COMPARISON_OPERATOR_INVERSES:
            with self.subTest(operator=operator):
                retry = {
                    'count': 30,
                    'delay': 5,
                    'continue-on': '<% $.foo {op} "continue" %>'.format(op=operator),
                }
                converter = WorkflowConverter()
                actual = converter.convert_retry(retry, 'retry_task_continue_on')
                inverse_operator = converter._get_inverse_comparison_operator(operator)
                expected = OrderedMap([
                    ('count', 30),
                    ('delay', 5),
                    ('when', '<% ctx().foo {op} "continue" %>'.format(op=inverse_operator)),
                ])
                self.assertEqual(expected, actual)

    def test_retry_break_on(self):
        # unittest2's alternative to @pytest.mark.parametrize
        # https://docs.python.org/3/library/unittest.html#distinguishing-test-iterations-using-subtests  # noqa: E501
        for operator in COMPARISON_OPERATOR_INVERSES:
            with self.subTest(operator=operator):
                retry = {
                    'count': 30,
                    'delay': 5,
                    'break-on': '<% $.bar {op} "break" %>'.format(op=operator),
                }
                converter = WorkflowConverter()
                actual = converter.convert_retry(retry, 'retry_task_break_on')
                inverse_operator = converter._get_inverse_comparison_operator(operator)
                expected = OrderedMap([
                    ('count', 30),
                    ('delay', 5),
                    ('when', '<% ctx().bar {op} "break" %>'.format(op=inverse_operator)),
                ])
                self.assertEqual(expected, actual)

    def test_retry_continue_and_break_on(self):
        retry = {
            'count': 30,
            'delay': 5,
            'continue-on': '<% $.foo = "continue" %>',
            'break-on': '<% $.foo = "break" %>',
        }
        converter = WorkflowConverter()
        with self.assertRaises(NotImplementedError):
            converter.convert_retry(retry, 'retry_task_continue_and_break_on')

    def test_convert_tasks(self):
        converter = WorkflowConverter()
        expr_converter = JinjaExpressionConverter()
        mistral_tasks = OrderedMap([
            ('jinja_task', OrderedMap([
                ('action', 'mypack.actionname'),
                ('input', OrderedMap([
                    ('cmd', '{{ _.test }}'),
                ])),
                ('on-success', ['next_task']),
            ]))
        ])
        result = converter.convert_tasks(mistral_tasks, expr_converter, set())
        self.assertEqual(result, OrderedMap([
            ('jinja_task', OrderedMap([
                ('action', 'mypack.actionname'),
                ('input', OrderedMap([
                    ('cmd', '{{ ctx().test }}'),
                ])),
                ('next', [
                    OrderedMap([
                        ('when', '{{ succeeded() }}'),
                        ('do', [
                            'next_task',
                        ])
                    ]),
                ]),
            ]))
        ]))

    def test_convert_tasks_yaql(self):
        converter = WorkflowConverter()
        expr_converter = YaqlExpressionConverter()
        mistral_tasks = OrderedMap([
            ('jinja_task', OrderedMap([
                ('action', 'mypack.actionname'),
                ('input', OrderedMap([
                    ('cmd', '<% $.test %>'),
                ])),
                ('on-success', ['next_task']),
            ]))
        ])
        result = converter.convert_tasks(mistral_tasks, expr_converter, set())
        self.assertEqual(result, OrderedMap([
            ('jinja_task', OrderedMap([
                ('action', 'mypack.actionname'),
                ('input', OrderedMap([
                    ('cmd', '<% ctx().test %>'),
                ])),
                ('next', [
                    OrderedMap([
                        ('when', '<% succeeded() %>'),
                        ('do', [
                            'next_task',
                        ])
                    ]),
                ]),
            ]))
        ]))

    def test_convert_tasks_join(self):
        converter = WorkflowConverter()
        expr_converter = YaqlExpressionConverter()
        mistral_tasks = OrderedMap([
            ('jinja_task', OrderedMap([
                ('action', 'mypack.actionname'),
                ('join', 'all'),
            ]))
        ])
        result = converter.convert_tasks(mistral_tasks, expr_converter, set())
        self.assertEqual(result, OrderedMap([
            ('jinja_task', OrderedMap([
                ('action', 'mypack.actionname'),
                ('join', 'all'),
            ]))
        ]))

    def _create_task(self, attr):
        return OrderedMap([
            ('bad_task', OrderedMap([
                ('action', 'mypack.actionname'),
                attr,
            ]))
        ])

    def test_convert_action_unsupported_mistral_builtin_actions(self):
        converter = WorkflowConverter()

        with self.assertRaises(NotImplementedError):
            converter.convert_action('std.echo')
        with self.assertRaises(NotImplementedError):
            converter.convert_action('std.email')
        with self.assertRaises(NotImplementedError):
            converter.convert_action('std.javascript')
        with self.assertRaises(NotImplementedError):
            converter.convert_action('std.js')
        with self.assertRaises(NotImplementedError):
            converter.convert_action('std.ssh')

    def test_convert_action_mistral_builtin_actions(self):
        converter = WorkflowConverter()

        self.assertEqual(converter.convert_action('std.fail'), 'fail')
        self.assertEqual(converter.convert_action('std.http'), 'core.http')
        self.assertEqual(converter.convert_action('std.mistral_http'), 'core.http')
        self.assertEqual(converter.convert_action('std.noop'), 'core.noop')

    def test_convert_action_others(self):
        converter = WorkflowConverter()

        # Test non-built-in actions
        self.assertEqual(converter.convert_action('basil.exposition'), 'basil.exposition')
        self.assertEqual(converter.convert_action('vanessakensington'), 'vanessakensington')
        self.assertEqual(converter.convert_action('foxxy_cleopatra'), 'foxxy_cleopatra')
        self.assertEqual(converter.convert_action('number 3'), 'number 3')

    def test_convert_tasks_unsupported_attributes(self):
        converter = WorkflowConverter()
        expr_converter = YaqlExpressionConverter()

        with self.assertRaises(NotImplementedError):
            mistral_tasks = self._create_task(('keep-result', True))
            converter.convert_tasks(mistral_tasks, expr_converter, set())

        with self.assertRaises(NotImplementedError):
            mistral_tasks = self._create_task(('pause-before', True))
            converter.convert_tasks(mistral_tasks, expr_converter, set())

        with self.assertRaises(NotImplementedError):
            mistral_tasks = self._create_task(('safe-rerun', True))
            converter.convert_tasks(mistral_tasks, expr_converter, set())

        with self.assertRaises(NotImplementedError):
            mistral_tasks = self._create_task(('target', 'some-node'))
            converter.convert_tasks(mistral_tasks, expr_converter, set())

        with self.assertRaises(NotImplementedError):
            mistral_tasks = self._create_task(('timeout', 60))
            converter.convert_tasks(mistral_tasks, expr_converter, set())

        with self.assertRaises(NotImplementedError):
            mistral_tasks = self._create_task(('wait-after', 60))
            converter.convert_tasks(mistral_tasks, expr_converter, set())

        with self.assertRaises(NotImplementedError):
            mistral_tasks = self._create_task(('wait-before', 60))
            converter.convert_tasks(mistral_tasks, expr_converter, set())

        with self.assertRaises(NotImplementedError):
            mistral_tasks = self._create_task(('workflow', 'someworkflowname'))
            converter.convert_tasks(mistral_tasks, expr_converter, set())

    def test_expr_type_converter_none(self):
        expr_type = None
        converter = WorkflowConverter()
        result = converter.expr_type_converter(expr_type)
        self.assertIsInstance(result, JinjaExpressionConverter)

    def test_expr_type_converter_string_jinja(self):
        expr_type = 'jinja'
        converter = WorkflowConverter()
        result = converter.expr_type_converter(expr_type)
        self.assertIsInstance(result, JinjaExpressionConverter)

    def test_expr_type_converter_string_yaql(self):
        expr_type = 'yaql'
        converter = WorkflowConverter()
        result = converter.expr_type_converter(expr_type)
        self.assertIsInstance(result, YaqlExpressionConverter)

    def test_expr_type_converter_string_bad_raises(self):
        expr_type = 'junk'
        converter = WorkflowConverter()
        with self.assertRaises(TypeError):
            converter.expr_type_converter(expr_type)

    def test_expr_type_converter_class_jinja(self):
        expr_type = JinjaExpressionConverter()
        converter = WorkflowConverter()
        result = converter.expr_type_converter(expr_type)
        self.assertIs(result, expr_type)

    def test_expr_type_converter_class_yaql(self):
        expr_type = YaqlExpressionConverter()
        converter = WorkflowConverter()
        result = converter.expr_type_converter(expr_type)
        self.assertIs(result, expr_type)

    def test_expr_type_converter_class_bad_raises(self):
        expr_type = []
        converter = WorkflowConverter()
        with self.assertRaises(TypeError):
            converter.expr_type_converter(expr_type)

    def test_convert_empty(self):
        mistral_wf = {}
        converter = WorkflowConverter()
        result = converter.convert(mistral_wf)
        self.assertEqual(result, {'version': '1.0'})

    def test_convert_all(self):
        mistral_wf = OrderedMap([
            ('description', 'test description'),
            ('type', 'direct'),
            ('input', [
                'data',
                {'jinja_expr': "{{ _.test }}"},
                {'yaql_expr': "<% $.test %>"},
            ]),
            ('vars', OrderedMap([
                ('var_data', "data"),
                ('var_expr_jinja', "{{ _.ex }}"),
                ('var_expr_yaql', "<% $.ex %>"),
            ])),
            ('output', OrderedMap([
                ('stdout', "{{ _.stdout }}"),
                ('stderr', "<% $.stderr %>"),
            ])),
            ('tasks', OrderedMap([
                ('jinja_task', OrderedMap([
                    ('action', 'mypack.actionname'),
                    ('input', OrderedMap([
                        ('cmd', '{{ _.test }}'),
                    ])),
                ])),
            ])),
        ])
        converter = WorkflowConverter()
        result = converter.convert(mistral_wf)

        self.assertEqual(result, OrderedMap([
            ('version', '1.0'),
            ('description', 'test description'),
            ('input', [
                'data',
                OrderedMap([
                    ('jinja_expr', "{{ ctx().test }}"),
                ]),
                OrderedMap([
                    ('yaql_expr', "<% ctx().test %>"),
                ]),
            ]),
            ('vars', [
                OrderedMap([
                    ('var_data', "data"),
                ]),
                OrderedMap([
                    ('var_expr_jinja', "{{ ctx().ex }}"),
                ]),
                OrderedMap([
                    ('var_expr_yaql', "<% ctx().ex %>"),
                ]),
            ]),
            ('output', [
                {'stdout': '{{ ctx().stdout }}'},
                {'stderr': '<% ctx().stderr %>'},
            ]),
            ('tasks', OrderedMap([
                ('jinja_task', OrderedMap([
                    ('action', 'mypack.actionname'),
                    ('input', OrderedMap([
                        ('cmd', '{{ ctx().test }}')
                    ])),
                ])),
            ])),
        ]))

    def test_convert_workflow_unsupported_attributes(self):
        mistral_wf = OrderedMap([
            ('version', '1.0'),
            ('output-on-error', OrderedMap([
                ('stdout', "{{ _.stdout }}"),
                ('stderr', "<% $.stderr %>"),
            ])),
        ])
        converter = WorkflowConverter()
        with self.assertRaises(NotImplementedError):
            converter.convert(mistral_wf)

    def test_convert_workflow_unsupported_types(self):
        converter = WorkflowConverter()
        with self.assertRaises(NotImplementedError):
            converter.convert(OrderedMap([
                ('version', '1.0'),
                ('type', 'reverse'),
            ]))

        with self.assertRaises(NotImplementedError):
            converter.convert(OrderedMap([
                ('version', '1.0'),
                ('type', 'junk'),
            ]))



# -*- coding: utf-8 -*-
# Licensed to the StackStorm, Inc ('StackStorm') under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os.path

from setuptools import setup, find_packages

from dist_utils import fetch_requirements
from dist_utils import apply_vagrant_workaround
from dist_utils import get_version_string

MODULE_NAME = 'orquestaconvert'
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
REQUIREMENTS_FILE = os.path.join(BASE_DIR, 'requirements.txt')
INIT_FILE = os.path.join(BASE_DIR, MODULE_NAME + '/__init__.py')

install_reqs, dep_links = fetch_requirements(REQUIREMENTS_FILE)

with open("README.md", "r") as fh:
    long_description = fh.read()

apply_vagrant_workaround()
setup(
    name=MODULE_NAME,
    version=get_version_string(INIT_FILE),
    description='Tool to convert OpenStack Mistral workflows to StackStorm Orquesta workflows',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Encore Technologies',
    author_email='code@encore.tech',
    url='https://encore.tech/',
    install_requires=install_reqs,
    dependency_links=dep_links,
    test_suite=MODULE_NAME,
    zip_safe=False,
    include_package_data=True,
    packages=find_packages(exclude=['setuptools', 'tests']),
    scripts=[
        'bin/orquestaconvert.sh',
        'bin/orquestaconvert-pack.sh',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Developers',
        'Environment :: Console',
    ],
)



from __future__ import print_function

import filecmp
import os
import six
import sys

from orquestaconvert.client import Client
from orquestaconvert.pack_client import PackClient

from tests.base_test_case import BasePackClientRunTestCase


class PackClientRunTestCase(BasePackClientRunTestCase):
    __test__ = True

    def setUp(self):
        super(PackClientRunTestCase, self).setUp()

        self.client = Client()
        self.pack_client = PackClient()

    def _validate_dirs(self, dir1, dir2):
        '''Make sure the directories are the same'''
        dirdiff = filecmp.dircmp(dir1, dir2)

        for diff_file in dirdiff.diff_files:
            sys.__stdout__.write(os.system("diff {} {}\n".format(
                os.path.join(self.m_actions_dir, diff_file),
                os.path.join(self.o_actions_dir, diff_file),
            )))

        self.assertEqual(dirdiff.diff_files, [])
        self.assertEqual(dirdiff.funny_files, [])

    def test_list_mistral_workflows_in_pack(self):
        args = ['--list-workflows=mistral-v2', '--actions-dir={}'.format(self.m_actions_dir)]
        self.pack_client.run(args, self.stdout, client=self.client)

        out = self.stdout.getvalue()

        self.assertEqual(len([line for line in out.split('\n') if line]), len(self.action_files))
        for action_file, wf_file in six.iteritems(self.action_wfs):
            self.assertIn('{action_file} --> {wf_file}\n'
                          .format(action_file=action_file, wf_file=wf_file),
                          out)
        self.assertEqual(self.stderr.getvalue(), '')

        self._validate_dirs(self.p_actions_dir, self.m_actions_dir)

    def test_list_orquesta_workflows_in_mistral_directory(self):
        args = ['--list-workflows=orquesta', '--actions-dir={}'.format(self.m_actions_dir)]
        self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(self.stdout.getvalue(), '')
        self.assertEqual(self.stderr.getvalue(), '')

        self._validate_dirs(self.p_actions_dir, self.m_actions_dir)

    def test_list_orquesta_workflows_in_pack(self):
        args = ['--list-workflows=orquesta', '--actions-dir={}'.format(self.o_actions_dir)]
        self.pack_client.run(args, self.stdout, client=self.client)

        out = self.stdout.getvalue()
        self.assertEqual(self.stderr.getvalue(), '')
        for action_file, wf_file in six.iteritems(self.orquesta_action_wfs):
            self.assertIn('{action_file} --> {wf_file}\n'
                          .format(action_file=action_file, wf_file=wf_file),
                          out)

    def test_partially_convert_pack(self):
        args = ['-e', 'yaql', '--actions-dir={}'.format(self.m_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        # Check the exit code
        self.assertEqual(1, result)

        # Check the stdout and stderr
        self.assertEqual(self.stdout.getvalue(), '')
        err = self.stderr.getvalue()
        self.assertIn("ERROR: Unable to convert all Mistral workflows.\n", err)
        self.assertIn(
            "Cannot convert \"retry\" spec with both a \"break-on\" and \"continue-on\" "
            "attributes in the \"test-error-undo-retry\" task.\n"
            "Affected files:\n"
            "  - {m_wfs_dir}/mistral-fail-retry-continue-and-break-on.yaml\n"
            "\n".format(m_wfs_dir=self.m_wfs_dir),
            err)

        # Check the failing actions failed conversion
        for action in self.action_failing_files:
            m_actual_action = self.get_fixture_content('pack/actions/{}'.format(action))
            m_expected_action = self.get_fixture_content('pack/pristine_actions/{}'.format(action))

            self.assertEqual(m_expected_action, m_actual_action)

            m_actual_wf = self.get_fixture_content('pack/actions/workflows/{}'.format(action))
            m_expected_wf = self.get_fixture_content('pack/pristine_actions/workflows/{}'
                                                     .format(action))

            self.assertEqual(m_expected_wf, m_actual_wf)

        # Check all of the remaining actions worked
        for o_action in self.action_passing_files:
            o_actual_action = self.get_fixture_content('pack/actions/{}'.format(o_action))
            o_expected_action = self.get_fixture_content('pack/o_actions/{}'.format(o_action))

            self.assertEqual(o_expected_action, o_actual_action)

            o_actual_wf = self.get_fixture_content('pack/actions/workflows/{}'.format(o_action))
            o_expected_wf = self.get_fixture_content('pack/o_actions/workflows/{}'.format(o_action))

            self.assertEqual(o_expected_wf, o_actual_wf)

        self._validate_dirs(self.m_actions_dir, self.o_actions_dir)

    def test_completely_convert_pack(self):
        for afile in self.action_failing_files:
            os.remove(os.path.join(self.m_actions_dir, afile))
            os.remove(os.path.join(self.m_wfs_dir, afile))

        args = ['-e', 'yaql', '--actions-dir={}'.format(self.m_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(0, result)

        self.assertEqual(self.stdout.getvalue(), '')
        self.assertEqual(self.stderr.getvalue(), '')

        for o_wf in self.action_passing_files:
            actual = self.get_fixture_content('pack/actions/workflows/{}'.format(o_wf))
            expected = self.get_fixture_content('pack/o_actions/workflows/{}'.format(o_wf))

            self.assertEqual(expected, actual)

        self._validate_dirs(self.m_actions_dir, self.o_actions_dir)

    def test_validate_nothing(self):
        args = ['--validate', '--actions-dir={}'.format(self.m_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(self.stdout.getvalue(), '')
        self.assertEqual(self.stderr.getvalue(), '')

        self.assertEqual(0, result)

        self._validate_dirs(self.p_actions_dir, self.m_actions_dir)

    def test_validate_verbose_nothing(self):
        args = ['--validate', '--verbose', '--actions-dir={}'.format(self.m_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(self.stdout.getvalue(), '')
        self.assertEqual(self.stderr.getvalue(), '')

        self.assertEqual(0, result)

        self._validate_dirs(self.p_actions_dir, self.m_actions_dir)

    def test_validate_orquesta(self):
        files = ['mistral-test-cancel.yaml', os.path.join('workflows', 'mistral-test-cancel.yaml')]
        before_dirhash = self._hash_directory(self.o_actions_dir, files)

        args = ['--validate', '--actions-dir={}'.format(self.o_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        after_dirhash = self._hash_directory(self.o_actions_dir, files)

        self.assertEqual(self.stdout.getvalue(), '')
        self.assertEqual(self.stderr.getvalue(), '')

        self.assertEqual(0, result)

        self.assertEqual(before_dirhash, after_dirhash)

    def test_validate_verbose_orquesta(self):
        files = ['mistral-test-cancel.yaml', os.path.join('workflows', 'mistral-test-cancel.yaml')]
        before_dirhash = self._hash_directory(self.o_actions_dir, files)

        args = ['--validate', '--verbose', '--actions-dir={}'.format(self.o_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        after_dirhash = self._hash_directory(self.o_actions_dir, files)

        self.assertEqual(0, result)

        self.assertEqual(self.stderr.getvalue(), '')

        out = self.stdout.getvalue()
        for action in self.action_passing_files:
            wf = os.path.join(self.o_wfs_dir, action)
            self.assertIn("Successfully validated workflow from {}\n".format(wf), out)

        self.assertEqual(before_dirhash, after_dirhash)

    def test_validate_partially_converted_pack(self):
        self.test_partially_convert_pack()

        self.setup_captures()

        args = ['--validate', '--actions-dir={}'.format(self.m_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(self.stdout.getvalue(), '')
        self.assertEqual(self.stderr.getvalue(), '')

        self.assertEqual(0, result)

        self._validate_dirs(self.m_actions_dir, self.o_actions_dir)

    def test_validate_verbose_partially_converted_pack(self):
        self.test_partially_convert_pack()

        self.setup_captures()

        args = ['--validate', '--verbose', '--actions-dir={}'.format(self.m_actions_dir)]
        result = self.pack_client.run(args, self.stdout, client=self.client)

        self.assertEqual(0, result)

        self.assertEqual(self.stderr.getvalue(), '')

        out = self.stdout.getvalue()
        for o_action in self.action_passing_files:
            o_wf = os.path.join(self.m_wfs_dir, o_action)
            self.assertIn("Successfully validated workflow from {}\n".format(o_wf), out)

        self._validate_dirs(self.m_actions_dir, self.o_actions_dir)



import ruamel.yaml.comments
import six
import warnings
import orquesta.expressions.base

from orquestaconvert.expressions.jinja import JinjaExpressionConverter
from orquestaconvert.expressions.yaql import YaqlExpressionConverter
from orquestaconvert.utils import type_utils


class ExpressionConverter(object):

    @classmethod
    def convert(cls, expr, **kwargs):
        if (isinstance(expr, type_utils.dict_types)):
            return cls.convert_dict(expr, **kwargs)
        elif isinstance(expr, list):
            return cls.convert_list(expr, **kwargs)
        elif isinstance(expr, six.string_types):
            return cls.convert_string(expr, **kwargs)
        elif isinstance(expr, bool):
            return expr
        elif expr is None:
            # Note: The only point to explicitly converting None is to skip
            #       emitting the warning. Otherwise we would remove this elif
            #       and let the function return the input. Also, ruamel doesn't
            #       explicitly serialize null to None, it just leaves the key
            #       blank (in other words, it "implicitly" serializes it).
            #       Example:
            #
            #           next:
            #             - when: ...
            #               publish:
            #                 - continue:
            #               ...
            #
            #       See this link for more information:
            #       https://bitbucket.org/ruamel/yaml/issues/169/roundtripdumper-dumps-null-values
            return None
        elif isinstance(expr, int):
            return expr
        else:
            warnings.warn("Could not recognize expression '{}'; results may not "
                          "be accurate.".format(expr), SyntaxWarning)
            return expr

    @classmethod
    def expression_type(cls, expr):
        for name, evaluator in six.iteritems(orquesta.expressions.base.get_evaluators()):
            if evaluator.has_expressions(str(expr)):
                return name
        return None

    @classmethod
    def get_converter(cls, expr):
        expr_type = cls.expression_type(expr)
        if expr_type == 'jinja':
            return JinjaExpressionConverter
        elif expr_type == 'yaql':
            return YaqlExpressionConverter
        return None

    @classmethod
    def unwrap_expression(cls, expr):
        converter = cls.get_converter(expr)
        if converter:
            return converter.unwrap_expression(expr)
        # this isn't a Jinja or YAQL expression, so return the raw string
        return expr

    @classmethod
    def convert_string(cls, expr, **kwargs):
        # - task('xxx').result -> result()
        #    if 'xxx' != current task name, error
        # - _. -> ctx().
        # - $. -> ctx().
        # - st2kv. -> st2kv('xxx')
        # others?
        converter = cls.get_converter(expr)
        if converter:
            return converter.convert_string(expr, **kwargs)
        # this isn't a Jinja or YAQL expression, so return the raw string
        return expr

    @classmethod
    def convert_dict(cls, expr_dict, **kwargs):
        converted = ruamel.yaml.comments.CommentedMap()
        for k, expr in six.iteritems(expr_dict):
            converted[cls.convert(k, **kwargs)] = cls.convert(expr, **kwargs)
        return converted

    @classmethod
    def convert_list(cls, expr_list, **kwargs):
        return [cls.convert(expr, **kwargs) for expr in expr_list]



from tests.base_test_case import BaseTestCase

from orquestaconvert.utils import yaml_utils

import ruamel.yaml
OrderedMap = ruamel.yaml.comments.CommentedMap


class TestYamlUtils(BaseTestCase):
    __test__ = True

    def test_yaml_to_obj(self):
        yaml_str = ("test_dict:\n"
                    "  a: b\n"
                    "\n"
                    "test_list:\n"
                    "  - x\n"
                    "  - y\n")
        result = yaml_utils.yaml_to_obj(yaml_str)
        self.assertEqual(result, OrderedMap([
            ('test_dict', OrderedMap([
                ('a', 'b')
            ])),
            ('test_list', ['x', 'y'])
        ]))

    def test_read_yaml(self):
        fixture_path = self.get_fixture_path('yaml/simple.yaml')
        data, ruamel_data = yaml_utils.read_yaml(fixture_path)
        self.assertEqual(data, {'test_dict': {'a': True}})
        self.assertEqual(data, OrderedMap([
            ('test_dict', OrderedMap([
                ('a', True)
            ]))
        ]))

    def test_obj_to_yaml(self):
        ruamel_data = OrderedMap([
            ('test_dict', OrderedMap([
                ('a', True)
            ]))
        ])
        result = yaml_utils.obj_to_yaml(ruamel_data)
        self.assertEqual(result, self.get_fixture_content('yaml/simple.yaml'))

    def test_obj_to_yaml_dict(self):
        ruamel_data = {'test_dict': {'a': True}}
        result = yaml_utils.obj_to_yaml(ruamel_data)
        self.assertEqual(result, self.get_fixture_content('yaml/simple.yaml'))

    def test_obj_to_yaml_no_line_wrap(self):
        ruamel_data = {
            'key': ("this is some super super super long line that would normally cause"
                    " a line wrap when converting from dict to yaml. instead we don't want"
                    " to do that and just keep this one crazy long line.")
        }
        result = yaml_utils.obj_to_yaml(ruamel_data)
        self.assertEqual(result, self.get_fixture_content('yaml/no_line_wrap.yaml'))
